#include "LuaPlus/LuaPlus.h"
using namespace LuaPlus;
#include "SimpleHeap.h"
#include "Timer.h"
#include <math.h>
#include <assert.h>

/**
	The LuaState class is able to make Lua callbacks look like natural extensions
	of the LuaState class.
**/
static int LS_PrintNumber(LuaState* state)
{
	LuaStack args(state);

	// Verify it is a number and print it.
	if (args[1].IsNumber())
		printf("%f\n", args[1].GetNumber());

	// No return values.
	return 0;
}


/**
**/
class ObjectWrapper
{
public:
	void Zero()
	{
	}

	void Print(const char* str)
	{
		printf("%s\n", str);
	}

	float Add(float num1, float num2)
	{
		return num1 + num2;
	}

	int LS_Add(LuaState* state)
	{
		LuaStack args(state);

		// Verify it is a number and print it.
		if (args[1].IsNumber()  &&  args[2].IsNumber())
		{
			state->PushNumber(args[1].GetNumber() + args[2].GetNumber() );
		}
		else
		{
			state->PushNumber( 0.0 );
		}

		// 1 return value.
		return 1;
	}
};


/**
**/
void DoScriptFormatTest()
{
	LuaStateOwner state;
	state->DoFile("ScriptVectorDump.lua");
}


int Test(int num)
{
	num;
	return 5;
}


/**
	Demonstrate registering callback functions for the Lua script.
**/
void DoScriptCallbackTest()
{
/*	LuaStateOwner state;

	state->GetGlobals().Register("PrintNumber", LS_PrintNumber);

	ObjectWrapper obj;
	state->GetGlobals().Register("Add", obj, ObjectWrapper::LS_Add);

	state->DoFile("ScriptCallbackTest.lua");
*/

	LuaStateOwner state;
/*
	state->GetGlobals().Register("PrintNumber", LS_PrintNumber);

	ObjectWrapper obj;
	state->GetGlobals().Register("Add", obj, ObjectWrapper::LS_Add);

	state->DoFile("ScriptCallbackTest.lua");

	// Now call it ourselves.
	LuaFunction<int> function(state, "PrintNumber");
	function(22);

	LuaFunction<float> add(state, "Add");
	float ret = add(22, 11);
*/
	LuaObject globalsObj = state->GetGlobals();

	LuaObject testTableObj = globalsObj.CreateTable("TestTable");
	testTableObj.Register("PrintNumber", LS_PrintNumber);
	state->DoString("TestTable.PrintNumber(1000)");

	// Test auto thing.
	int top = state->GetTop();
	globalsObj.RegisterDirect("Test", Test);
	int top2 = state->GetTop();
	globalsObj.RegisterDirect("sin", (double (*)(double))&sin);
	LuaFunction<int> funcCall(state, "Test");
	int value = funcCall(10);

	LuaFunction<float> funcCall2(state, "sin");
	float value2 = funcCall2(5);

	ObjectWrapper obj;
	globalsObj.RegisterDirect("Add", obj, ObjectWrapper::Add);
	globalsObj.RegisterDirect("Zero", obj, ObjectWrapper::Zero);
	globalsObj.RegisterDirect("Print", obj, ObjectWrapper::Print);
	LuaFunction<double> addCall(state, "Add");
	LuaFunctionVoid printCall(state, "Print");
	LuaFunctionVoid zeroCall(state, "Zero");
	zeroCall();
	printCall("Stuff");
	double result = addCall(10, 2);

}


/**
	Demonstrate reading and saving a script.
**/
void DoScriptSaveTest()
{
	LuaStateOwner state;

	state->DoFile( "ScriptSaveTest.lua" );
//	state->DumpGlobalsFile("ScriptSaveTest.dmp");
}


/**
	Demonstrates walking an array table.
**/
void DoScriptArrayTest()
{
	LuaStateOwner state;
	if (state->DoFile("ScriptArrayTest.lua") != 0)
		printf("Unable to read ScriptArrayTest.lua.\n");
	LuaObject testTableObj = state->GetGlobals()[ "TestArray" ];
	for (int i = 1; ; ++i)
	{
		LuaObject entryObj = testTableObj[ i ];
		if (entryObj.IsNil())
			break;
		if (entryObj.IsNumber())
			printf("%f\n", entryObj.GetNumber());
		else if (entryObj.IsString())
			printf("%s\n", entryObj.GetString());
	}
}


static int LS_LightUserDataCall( LuaState* state )
{
	LuaStack args(state);
	bool isLightUserData = args[ 1 ].IsLightUserData();
	const void* ptr = args[ 1 ].GetUserData();
	return 0;
}


void TestPointer()
{
	LuaStateOwner state;

	state->GetGlobals().Register("LightUserDataCall", LS_LightUserDataCall);
	LuaObject funcObj = state->GetGlobal("LightUserDataCall");
	LuaCall call = funcObj;
	call << (void*)0xfedcba98 << LuaRun();
}


/*void LuaStackTableIteratorTest()
{
	LuaStateOwner state;

	state->DoString( "Table = { Hi = 5, Hello = 10, Yo = 6 }" );

	int origTop = state->GetTop();

	LuaStackObject obj = state->GetGlobal("Table");

	for ( LuaStackTableIterator it( obj ); it; it.Next() )
	{
		const char* key = it.GetKey().GetString();
		int num = it.GetValue().GetInteger();
	}

	LuaStackObject obj2 = state->GetGlobal("Table");

	for ( it.Reset(); it; ++it )
	{
		const char* key = it.GetKey().GetString();
	}
}
*/

void LuaTableIteratorTest()
{
	LuaStateOwner state;

	state->DoString( "Table = { Hi = 5, Hello = 10, Yo = 6 }" );

	int origTop = state->GetTop();

	LuaObject obj = state->GetGlobal("Table");

	for ( LuaTableIterator it( obj ); it; it.Next() )
	{
		const char* key = it.GetKey().GetString();
		int num = it.GetValue().GetInteger();
	}
}


void TestNewCall()
{
	LuaStateOwner state;

	state->DoString("function Add(x, y) return x + y end");
	LuaObject funcObj = state->GetGlobal("Add");
	int top = state->GetTop();
	LuaCall call = funcObj;
	LuaObject retObj = call << 2 << 7 << LuaRun();
	int top2 = state->GetTop();
//	funcObj() << LuaRun();
}

using namespace SimpleMem;

SimpleHeap* heap;

void* ReallocFunction(void* ptr, int oldsize, int size, void* data, const char* allocName, unsigned int flags)
{
	SimpleHeap* heap = (SimpleHeap*)data;

	if (allocName  &&  strncmp(allocName, "Lua_parser", 10) == 0)
	{
//		int hi = 5;
	}

	if (ptr != NULL)
	{
		int hi = 5;
	}

//	MemoryManager::Info* info = MemoryManager::GetInfo();
//	info->m_curAllocName = allocName;
	if (allocName == NULL)
	{
		int hi = 5;	hi;
	}
	BYTE* ret;

	if (flags == LUA_ALLOC_TEMP)
	{
		ret = (BYTE*)heap->Realloc(ptr, size, SimpleHeap::FIRSTFIT_TOP,
			allocName, __FILE__, __LINE__);
	}
	else
	{
		ret = (BYTE*)heap->Realloc(ptr, size, SimpleHeap::FIRSTFIT_BOTTOM,
			allocName, __FILE__, __LINE__);
	}
	return ret;
}


void FreeFunction(void* ptr, int oldsize, void* data)
{
	SimpleHeap* heap = (SimpleHeap*)data;
	heap->Free(ptr);
}


int checkpoint = 0;

class MemLogFile
{
public:
	MemLogFile()
	{
		Create("c:\\MemoryDump.txt");
	}

	~MemLogFile()
	{
		Close();
	}

	void Close()
	{
		if (m_file)
		{
			fclose(m_file);
			m_file = NULL;
		}
	}

	void Create(const char* fileName)
	{
		Close();

		strcpy(m_fileName, fileName);
		m_file = fopen(m_fileName, "wt");
		assert(m_file);
		Write("Heap Name\tBlock Type\tAddress\tRequested Size\tTotal Size\tIndex\tName\tFile Name\tLine Number\n");
	}

	void Open()
	{
		Close();
		m_file = fopen(m_fileName, "at");
		assert(m_file);
	}

	void Write(const char* msg, ...)
	{
		va_list args;
		va_start(args, msg);
		vfprintf(m_file, msg, args);
		va_end(args);
	}

	void Flush()
	{
		fflush(m_file);
	}

	operator FILE*()				{  return m_file;  }

protected:
	FILE* m_file;
	char m_fileName[_MAX_PATH];
};


MemLogFile& GetMemLogFile()
{
	static MemLogFile autoFile;
	return autoFile;
}


void DumpIt()
{
	heap->HeapDump(GetMemLogFile(), checkpoint, checkpoint);
}


int Chunkwriter(lua_State *L, const void* p,
                size_t sz, void* ud)
{
	FILE* file = (FILE*)ud;
	fwrite(p, sz, 1, file);
	return sz;
}


void TestHeap()
{
	SimpleHeap mainHeap("Test");
	mainHeap.Initialize(1 * 1024 * 1024);
	heap = &mainHeap;

	GetMemLogFile().Create("c:\\dump.txt");

	heap->SetAlignment(1);

	heap->Alloc(1);
	DumpIt();
	void* p0 = heap->Alloc(1);
	DumpIt();
	void* p1 = heap->Alloc(1);
	DumpIt();
	heap->Free(p0);
	DumpIt();
	heap->Free(p1);
	DumpIt();

	heap->Alloc(100);
	DumpIt();
	heap->Alloc(100);
	DumpIt();
	heap->Alloc(100);
	DumpIt();
	heap->Alloc(100, SimpleHeap::FIRSTFIT_TOP);
	DumpIt();
	heap->Alloc(100, SimpleHeap::FIRSTFIT_TOP);
	DumpIt();
	heap->Free(p1);
	DumpIt();
	heap->Alloc(50);
	DumpIt();
	heap->Alloc(50);
	DumpIt();
	heap->Alloc(1047300);
	DumpIt();
	heap->Alloc(2, SimpleHeap::FIRSTFIT_TOP);
	DumpIt();
}


void TestGCObject()
{
	TestHeap();
	SimpleHeap mainHeap("Test");
	mainHeap.Initialize(1 * 1024 * 1024);
	heap = &mainHeap;

	GetMemLogFile().Create("c:\\dump.txt");

	lua_setdefaultmemoryfunctions(ReallocFunction, FreeFunction, &mainHeap);
	LuaStateOwner state(false);
	lua_setdefaultmemoryfunctions(NULL, NULL, NULL);
	state->CollectGarbage();

	heap->LogStats();
	DumpIt();

//	checkpoint = heap->SetCheckpoint();

/*	{
		LuaObject stringObj(state);
		stringObj.AssignString("Hello, world!");
		LuaObject globalsObj = state->GetGlobals();
	}

	state->CollectGarbage();
*/
//	state->DoString("Table = { 0, 1, 2, 'Hello', nil, 'Hi', Yo = 'My Stuff', NobodysHome = 5, NestedTable = { 1, 2, 3, { 'String', }, { 'Table2' } }, { 'String1' } }");
//	state->DoString("a = 5");
/*	FILE* inFile = fopen("c:\\dump.dat", "rb");
	fseek(inFile, 0, SEEK_END);
	int size = ftell(inFile);
	fseek(inFile, 0, SEEK_SET);
	BYTE* buf = new BYTE[size];
	fread(buf, size, 1, inFile);
	fclose(inFile);

	state->DoBuffer((const char*)buf, size, NULL);
	state->CollectGarbage();
*/
	LPCSTR strbuf = "a = 5";
//	state->CheckStack(1);
//	state->LoadBuffer(strbuf, strlen(strbuf), "Stuff");
	state->DoString(strbuf);
/*  FILE* file = fopen("c:\\dump.dat", "wb");
	state->CheckStack(1);
	state->Dump(Chunkwriter, file);
	fclose(file);*/
	state->CollectGarbage();

	{
		LuaObject obj;

	}
}


void SetUserData()
{
	SimpleHeap mainHeap("Test");
	mainHeap.Initialize(1 * 1024 * 1024);
	heap = &mainHeap;

	GetMemLogFile().Create("c:\\dump.txt");

	{
		lua_setdefaultmemoryfunctions(ReallocFunction, FreeFunction, &mainHeap);
		LuaStateOwner state(false);
		lua_setdefaultmemoryfunctions(NULL, NULL, NULL);
		{
			LuaObject obj(state);
			obj.AssignNewTable(state);
			assert(obj.IsTable());

			// number
			obj.SetUserData(1, (void*)0x12345678);
			assert(obj[1].GetUserData() == (void*)0x12345678);

			// string
			obj.SetUserData("Hello", (void*)0x87654321);
			assert(obj["Hello"].GetUserData() == (void*)0x87654321);

			// object
			LuaObject stringObj(state);
			stringObj.AssignString(state, "Test");
			obj.SetUserData(stringObj, (void*)0x02468024);
			assert(obj[stringObj].GetUserData() == (void*)0x02468024);
		}
		state->CollectGarbage();
		DumpIt();
	}
}


void CloneTest()
{
	SimpleHeap mainHeap("Test");
	mainHeap.Initialize(1 * 1024 * 1024);
	heap = &mainHeap;

	GetMemLogFile().Create("c:\\dump.txt");

	lua_setdefaultmemoryfunctions(ReallocFunction, FreeFunction, &mainHeap);
	LuaStateOwner state;
	lua_setdefaultmemoryfunctions(NULL, NULL, NULL);

	LuaObject valueObj(state);
	valueObj.AssignBoolean(state, true);

	LuaObject cloneObj = valueObj.Clone();

	Timer timer;
	timer.Start();
	state->DoString("Table = { 0, 1, 2, 'Hello', nil, 'Hi', Yo = 'My Stuff', NobodysHome = 5, NestedTable = { 1, 2, 3, { 'String', }, { 'Table2' } }, { 'String1' } }");
	timer.Stop();
	printf("DoString: %f\n", timer.GetMillisecs());

	LuaObject tableObj = state->GetGlobal("Table");
	timer.Reset();
	timer.Start();
	LuaObject clonedTableObj = tableObj.Clone();
	timer.Stop();
	printf("Clone: %f\n", timer.GetMillisecs());

	clonedTableObj.SetNil("Yo");

	state->DumpObject("c:\\test1.lua", "Table", tableObj, false);
	state->DumpObject("c:\\test2.lua", "Table", clonedTableObj, false);
}


int CallbackA(LuaState* state)
{
	for (int i = 0; i < 10; ++i)
	{
		printf("Hello%d\n", i);
//		state->CoYield(0);
	}

	return 0;
}


int CallbackB(LuaState* state)
{
	for (int i = 0; i < 15; ++i)
	{
		printf("Hi%d\n", i);
//		state->CoYield(0);
	}

	return 0;
}


void TestThreeThreads()
{
	LuaStateOwner state;
	LuaStateAuto thread1 = LuaState::CreateThread(state);
	LuaStateAuto thread2 = LuaState::CreateThread(state);

	LuaObject obj(thread1);
	obj.AssignNewTable(state);

	LuaObject obj2(thread2);
	obj2.AssignString(state, "Hi");

	LuaObject obj3(thread2);
	obj3.AssignWString(state, L"Hello");

	state->GetGlobals().Register("CallbackA", CallbackA);
	state->GetGlobals().Register("CallbackB", CallbackB);

	LuaFunctionVoid callbackA = thread1->GetGlobal("CallbackA");
	LuaFunctionVoid callbackB = thread2->GetGlobal("CallbackB");

	callbackA();
	callbackB();

	state->CollectGarbage();
}


void TestDumpGlobals()
{
	LuaStateOwner state(false);
	state->DoString("GlobalTable = { 1, 2, 3, 4 }");
	state->DoString("GlobalValue = 5");
	state->DumpGlobals("c:\\dump.lua");
}


void DumpTest()
{
	LuaStateOwner state(false);
	LuaObject complexObj = state->GetGlobals().CreateTable("Complex");
	complexObj.SetString("d:\\Test\\Stuff\\\xff\xfe", "An entry");
	state->DumpObject("c:\\test.lua", "Complex", complexObj);
}


void TestANSIFile()
{
	LuaStateOwner state(true);
	state->DoFile("TestANSI.lua");

	LuaObject sObj = state->GetGlobal("s");
	const char* str = sObj.GetString();
}


void TestUnicodeFile()
{
	LuaStateOwner state(true);
	state->DoFile("TestUnicode.lua");

	LuaObject sObj = state->GetGlobal("s");
	const lua_WChar* str = sObj.GetWString();
}


void IntegerTest()
{
	LuaStateOwner state(true);
	state->DoString("i1 = 10");
	state->DoString("i2 = 20; f1 = 15.5; f2 = .5");
	state->DoString("i3 = -10");
	state->DoString("print(type(i1), type(i2), type(i3), type(f1), type(f2))");
	state->DoString("res1 = i1 + i2");
	state->DoString("print(type(res1), res1)");
	state->DoString("res1 = i1 + f1");
	state->DoString("print(type(res1), res1)");
	state->DoString("res = -i1");
	state->DoString("print(type(res), res)");
	state->DoString("print(i1 | 3)");
	state->DoString("print(i1 & 2)");
	state->DoString("print(i1 ^ 12)");
	state->DoString("print(1 << 4)");
	state->DoString("print(32 >> 4)");

	state->DoString("i1 = 0x1000");
	state->DoString("print(i1, type(i1))");
	state->DoString("print(i1 | 0x80000000)");

	state->DoString("for index = 1, 10, 0.5 do print(index) end");
	state->DoString("for index = 1, 10, 2 do print(index) end");
}


/*void LookupTest()
{
	LuaStateOwner state;
	state->DoString("s = 'Hello'  for key, value in s do print(value) end");
}
*/

class MultiObject
{
public:
	MultiObject(int num) :
		m_num(num)
	{
	}

	int Print(LuaState* state)
	{
		printf("%d\n", m_num);
		return 0;
	}

	void Print2(int num)
	{
		printf("%d %d\n", m_num, num);
	}

protected:
	int m_num;
};


void MultiObjectTest()
{
	LuaStateOwner state;

	LuaObject metaTableObj = state->GetGlobals().CreateTable("MultiObjectMetaTable");
	metaTableObj.SetObject("__index", metaTableObj);
	metaTableObj.RegisterObjectFunctor("Print", &MultiObject::Print);
	metaTableObj.RegisterObjectDirect("Print2", (MultiObject*)0, &MultiObject::Print2);

	MultiObject obj1(10);
	LuaObject obj1Obj = state->BoxPointer(&obj1);
	obj1Obj.SetMetaTable(metaTableObj);
	state->GetGlobals().SetObject("obj1", obj1Obj);

	MultiObject obj2(20);
	LuaObject obj2Obj = state->BoxPointer(&obj2);
	obj2Obj.SetMetaTable(metaTableObj);
	state->GetGlobals().SetObject("obj2", obj2Obj);

	state->DoString("obj1:Print()");
	state->DoString("obj2:Print()");
	state->DoString("obj1:Print2(5)");
	state->DoString("obj2:Print2(15)");

	LuaObject table1Obj = state->GetGlobals().CreateTable("table1");
	table1Obj.SetLightUserData("__object", &obj1);
	table1Obj.SetMetaTable(metaTableObj);

	LuaObject table2Obj = state->GetGlobals().CreateTable("table2");
	table2Obj.SetLightUserData("__object", &obj2);
	table2Obj.SetMetaTable(metaTableObj);

	state->DoString("table1:Print()");
	state->DoString("table2:Print()");
	state->DoString("table1:Print2(5)");
	state->DoString("table2:Print2(15)");
}


class lua_StateObjectHelper
{
public:
	lua_StateObjectHelper() :
		m_str("Me")
	{
	}

	int PrintSomething(lua_State* L)
	{
		printf("Hello, world! %s\n", m_str);
		return 0;
	}

protected:
	const char* m_str;
};


int PrintSomethingGlobal(lua_State* L)
{
	printf("In global PrintSomething.\n");
	return 0;
}


void lua_StateCallbackTest()
{
	lua_State* L = lua_open();

	lua_StateObjectHelper helper;

	lua_pushstring(L, "PrintSomething");
	lua_pushfunctorclosure(L, helper, lua_StateObjectHelper::PrintSomething, 0);
	lua_settable(L, LUA_GLOBALSINDEX);

	lua_pushstring(L, "PrintSomethingGlobal");
	lua_pushfunctorclosure(L, PrintSomethingGlobal, 0);
	lua_settable(L, LUA_GLOBALSINDEX);

	lua_dostring(L, "PrintSomething()");
	lua_dostring(L, "PrintSomethingGlobal()");

	lua_close(L);
}


void MemoryTest()
{
	SimpleHeap mainHeap("Test");
	mainHeap.Initialize(1 * 1024 * 1024);
	heap = &mainHeap;

	GetMemLogFile().Create("c:\\dump.txt");

	{
		lua_setdefaultmemoryfunctions(ReallocFunction, FreeFunction, &mainHeap);
		LuaStateOwner state(true);
		lua_setdefaultmemoryfunctions(NULL, NULL, NULL);
		{
//			state->DoFile("s:\\svndb\\lua-5.0.1\\test\\sieve.lua");
			DumpIt();
		}
		state->CollectGarbage();
		DumpIt();
	}
}


void RCTest()
{
	LuaStateOwner state(false);
	LuaObject stringObj(state);
	stringObj.AssignString(state, "Test");
	stringObj.AssignWString(state, L"Wide String");
}


void ExceptionTest()
{
	LuaStateOwner state;
	state->DoString("MyTable = { 1, 2, 3, 4 }");

	try
	{
		void* data = state->GetGlobal("MyTable")[1].GetUserData();
	}
	catch (const LuaException& e)
	{
		int hi = 5; (void)hi;
	}

	try
	{
		LuaObject obj = state->GetGlobals()["MyTable"][1][2];
	}
	catch (const LuaException& e)
	{
		int hi = 5; (void)hi;
	}
}


void BisectTest()
{
	LuaStateOwner state;
	state->DoFile("../../Test/bisect.lua");
}


#if _WIN32_WCE
int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPTSTR lpCmdLine, int nCmdShow)
#else _WIN32_WCE
int __cdecl main(int argc, char* argv[])
#endif _WIN32_WCE
{
	BisectTest();
	RCTest();
	ExceptionTest();

	int i;

	{
		LuaStateOwner state(false);
		state->DoString("s = \"12345678901234567890123456789012345678901234567890\"");
		state->DoString("s2 = L\"abcdefghijklmnopqrstuvwxyzzyxwvutsrqponmlkjihgfedcba\"");
	}
//	LookupTest();
	MemoryTest();
	lua_StateCallbackTest();
	MultiObjectTest();
	IntegerTest();

	LuaState* state = LuaState::Create(true);

	DWORD count = GetTickCount();
	lua_dostring(*state, "a = true;  if a then print(\"Hi\") end");

	lua_dostring(*state, "theGlobal = 5; function ReturnTrue() if theGlobal == 5 then return true else return false end end");

	LuaObject obj = state->GetGlobals().CreateTable("Stuff");
	for (i = 0; i < 100000; ++i)
	{
		obj.CreateTable(i);
	}

		lua_getfenv(*state, LUA_GLOBALSINDEX);
		lua_pushstring(*state, "ReturnTrue");
		lua_gettable(*state, -2);
	for (i = 0; i < 1000000; ++i)
	{
		lua_pushvalue(*state, -1);
//		char s[80];
//		sprintf(s, "%d", i);
//		lua_pushstring(*state, s);
		lua_pushnumber(*state, i);
		lua_call(*state, 1, 1);
		lua_pop(*state, 1);
	}

	state->CollectGarbage();

	count = GetTickCount() - count;
	printf("%d\n", count);


	TestANSIFile();
	TestUnicodeFile();
	DumpTest();
	TestDumpGlobals();
	SetUserData();
	TestGCObject();
	TestThreeThreads();
	TestNewCall();
	CloneTest();

	TestPointer();
	{
		LuaStateOwner state;
		state->DoString("Table = { 0, 1, 2, 'Hello', nil, 'Hi', Yo = 'My Stuff', NobodysHome = 5, NestedTable = { 1, 2, 3, { 'String', }, { 'Table2' } }, { 'String1' } }");
		LuaObject globalsObj = state->GetGlobals();
		LuaObject tableObj = state->GetGlobal("Table");
		LuaObject numObj = tableObj[2];
	}
	LuaTableIteratorTest();
//	LuaStackTableIteratorTest();

	DoScriptFormatTest();

	DoScriptCallbackTest();
	DoScriptSaveTest();
	DoScriptArrayTest();
	TestPointer();

	return 0;
}
