#pragma once

class SimpleString
{
	friend bool	operator== (const SimpleString& left, const SimpleString& right);

public:
						SimpleString ();
						SimpleString (const char *value);
						SimpleString (const SimpleString& other);
						~SimpleString ();

	SimpleString		operator= (const SimpleString& other);

	char				*asCharString () const;
	int					size() const;

private:
	char				*buffer;
};



SimpleString StringFrom (bool value);
SimpleString StringFrom (const char *value);
SimpleString StringFrom (long value);
SimpleString StringFrom (double value);
SimpleString StringFrom (const SimpleString& other);



//////////////////////////////////////


class TestResult;



class Test
{
public:
	Test (const SimpleString& testName);

	virtual void	run (TestResult& result) = 0;


	void			setNext(Test *test);
	Test			*getNext () const;

protected:

	bool check (long expected, long actual, TestResult& result, const SimpleString& fileName, long lineNumber);
	bool check (const SimpleString& expected, const SimpleString& actual, TestResult& result, const SimpleString& fileName, long lineNumber);

	SimpleString	name_;
	Test			*next_;

};


#define TEST(testName, testGroup)\
  class testGroup##testName##Test : public Test \
	{ public: testGroup##testName##Test () : Test (#testName "Test") {} \
            void run (TestResult& result_); } \
    testGroup##testName##Instance; \
	void testGroup##testName##Test::run (TestResult& result_) 



#define CHECK(condition)\
{ if (!(condition)) \
{ result_.addFailure (Failure (name_, __FILE__,__LINE__, #condition)); return; } }



#define CHECK_EQUAL(expected,actual)\
{ if ((expected) == (actual)) return; result_.addFailure(Failure(name_, __FILE__, __LINE__, StringFrom(expected), StringFrom(actual))); }


#define LONGS_EQUAL(expected,actual)\
{ long actualTemp = actual; \
  long expectedTemp = expected; \
  if ((expectedTemp) != (actualTemp)) \
{ result_.addFailure (Failure (name_, __FILE__, __LINE__, StringFrom(expectedTemp), \
StringFrom(actualTemp))); return; } }



#define DOUBLES_EQUAL(expected,actual,threshold)\
{ double actualTemp = actual; \
  double expectedTemp = expected; \
  if (fabs ((expectedTemp)-(actualTemp)) > threshold) \
{ result_.addFailure (Failure (name_, __FILE__, __LINE__, \
StringFrom((double)expectedTemp), StringFrom((double)actualTemp))); return; } }



#define FAIL(text) \
{ result_.addFailure (Failure (name_, __FILE__, __LINE__,(text))); return; }



///////////////////////////////////



class Failure;

class TestResult
{
public:
					TestResult ();
	virtual void	testsStarted ();
	virtual void	addFailure (const Failure& failure);
	virtual void	testsEnded ();

private:
	int				failureCount;
};



////////////////////////////////////


class Failure
{

public:
	Failure (const SimpleString&		theTestName,
			 const SimpleString&		theFileName,
			 long	  					theLineNumber,
			 const SimpleString&		theCondition);

	Failure (const SimpleString&		theTestName,
			 const SimpleString&		theFileName,
			 long						theLineNumber,
			 const SimpleString&		expected,
			 const SimpleString&		actual);

	SimpleString		message;
	SimpleString		testName;
	SimpleString		fileName;
	long				lineNumber;
};


////////////////////////////////////


class Test;
class TestResult;



class TestRegistry
{
public:
	static void addTest (Test *test);
	static void runAllTests (TestResult& result);

private:

	static TestRegistry&	instance ();
	void					add (Test *test);
	void					run (TestResult& result);


	Test					*tests;

};



/////////////////////////////////////


