// This is the main DLL file.

#include "stdafx.h"

#include "ManagedLuaPlus.h"
#include <crtdbg.h>
#include "_vcclrit.h"

#pragma warning(disable: 4239)

main()
{
}

namespace ManagedLuaPlus {

using namespace System::Reflection;
using System::Runtime::InteropServices::GCHandle;
using namespace System::Collections;

public __gc class LuaState;
public __gc class LuaStackObject;
public __gc class LuaObject;

public __delegate int LuaPlusCallback(LuaState* state);

class AString
{
public:
	AString(const AString& src) :
		 m_buffer(NULL)
	{
		Duplicate(src.m_buffer);
	}

	AString(const char* str) :
		m_buffer(NULL)
	{
		Duplicate(str);
	}

	~AString()
	{
		if (m_buffer)
			free(m_buffer);
	}

	operator const char*() const
	{
		return m_buffer;
	}

	AString& operator=(const AString& src)
	{
		Duplicate(src.m_buffer);
	}

protected:
	void Duplicate(const char* str)
	{
		if (m_buffer)
			free(m_buffer);
		m_buffer = (char*)malloc(strlen(str) + 1);
		strcpy(m_buffer, str);
	}

	char* m_buffer;
};



class WString
{
public:
	WString(const WString& src) :
		 m_buffer(NULL)
	{
		Duplicate(src.m_buffer);
	}

	WString(const wchar_t* str) :
		m_buffer(NULL)
	{
		Duplicate(str);
	}

	~WString()
	{
		delete[] m_buffer;
	}

	operator const wchar_t*() const
	{
		return m_buffer;
	}

	WString& operator=(const WString& src)
	{
		Duplicate(src.m_buffer);
	}

protected:
	void Duplicate(const wchar_t* str)
	{
		delete[] m_buffer;
		m_buffer = new wchar_t[wcslen(str) + 1];
		wcscpy(m_buffer, str);
	}

	wchar_t* m_buffer;
};



using namespace System::Runtime::InteropServices;

class MarshalAllocatorANSI
{
public:
	MarshalAllocatorANSI(System::String* s)
	{
		m_string = (const char*)(Marshal::StringToHGlobalAnsi(s)).ToPointer();
	}

	~MarshalAllocatorANSI()
	{
		Marshal::FreeHGlobal(IntPtr((void*)m_string));
	}

	operator const char*() const
	{
		return m_string;
	}

protected:
	const char* m_string;
};


class MarshalAllocatorUnicode
{
public:
	MarshalAllocatorUnicode(System::String* s)
	{
		m_string = (const wchar_t*)(Marshal::StringToHGlobalUni(s)).ToPointer();
	}

	~MarshalAllocatorUnicode()
	{
		Marshal::FreeHGlobal(IntPtr((void*)m_string));
	}

	operator const wchar_t*() const
	{
		return m_string;
	}

protected:
	const wchar_t* m_string;
};


// This is so inefficient it hurts.
AString MarshalString (System::String* s)
{
	MarshalAllocatorANSI alloc(s);
	return AString(alloc);
}

WString MarshalWString (System::String* s)
{
	MarshalAllocatorUnicode alloc(s);
	return WString(alloc);
}


public __gc class LuaNode
{
public:
	LuaNode(LuaObject* key, LuaObject* value) :
		m_key(key),
		m_value(value)
	{
	}

	__property LuaObject* get_Key()
	{
		return m_key;
	}

	__property LuaObject* get_Value()
	{
		return m_value;
	}

	String* ToString();

private:
	LuaObject* m_key;
	LuaObject* m_value;
};


public __gc class LuaException : public Exception
{
public:
	LuaException(String* message) :
		m_message(message)
	{
	}

	__property String* get_Message()
	{
		return m_message;
	}

private:
	String* m_message;
};


[System::Reflection::DefaultMemberAttribute(S"Item")]
public __gc class LuaObject : public IEnumerable
{
public:
	LuaObject() : m_object(new LuaPlus::LuaObject) {}
	LuaObject(ManagedLuaPlus::LuaObject* object);
	LuaObject(ManagedLuaPlus::LuaState* state, LuaPlus::LuaObject& object);
	LuaObject(ManagedLuaPlus::LuaState* state);
	LuaObject(ManagedLuaPlus::LuaState* state, int stackIndex);
	LuaObject(ManagedLuaPlus::LuaStackObject* object);

	static LuaObject* op_Assign(LuaObject* left, LuaObject* right)
	{
		delete left->m_object;
		left->m_parent = right->m_parent;
		left->m_object = new LuaPlus::LuaObject(*right->m_object);
		return left;
	}

	~LuaObject()
	{
		delete m_object;
	}

	/**
		Retrieves the LuaState object associated with this LuaStackObject.
	**/
	LuaState* GetState()				{  return m_parent;  }

	static int op_Equality(LuaObject* left, LuaObject* right)
	{
		return *left->m_object == *right->m_object;
	}

	static int op_LessThan(LuaObject* left, LuaObject* right)
	{
		return *left->m_object < *right->m_object;
	}

	String* TypeName()					{  return new String(m_object->TypeName());  }
	int Type()							{  return m_object->Type();  }

	bool IsNil()						{  return m_object->IsNil();  }
	bool IsTable()						{  return m_object->IsTable();  }
	bool IsUserData()					{  return m_object->IsUserData();  }
	bool IsCFunction()					{  return m_object->IsCFunction();  }
	bool IsNumber()						{  return m_object->IsNumber();  }
	bool IsString()						{  return m_object->IsString();  }
	bool IsWString()					{  return m_object->IsWString();  }
	bool IsConvertibleToNumber()		{  return m_object->IsConvertibleToNumber();  }
	bool IsConvertibleToString()		{  return m_object->IsConvertibleToString();  }
	bool IsConvertibleToWString()		{  return m_object->IsConvertibleToWString();  }
	bool IsFunction()					{  return m_object->IsFunction();  }
	bool IsNone()						{  return m_object->IsNone();  }
	bool IsLightUserData()				{  return m_object->IsLightUserData();  }
	bool IsBoolean()					{  return m_object->IsBoolean();  }

	double ToNumber()					{  return m_object->ToNumber();  }
	String* ToString()					{  return new String(m_object->ToString());  }
	String* ToWString()					{  return new String(m_object->ToWString());  }
	size_t ToStrLen()					{  return m_object->ToStrLen();  }

	int GetInteger()
	{
		try
		{
			return m_object->GetInteger();
		}
		catch (const LuaPlus::LuaException& e)
		{
			throw new LuaException(new String(e.GetMessage()));
		}
	}

	float GetFloat()
	{
		try
		{
			return m_object->GetFloat();
		}
		catch (const LuaPlus::LuaException& e)
		{
			throw new LuaException(new String(e.GetMessage()));
		}
	}

	double GetNumber()
	{
		try
		{
			return m_object->GetNumber();
		}
		catch (const LuaPlus::LuaException& e)
		{
			throw new LuaException(new String(e.GetMessage()));
		}
	}

	String* GetString()
	{
		try
		{
			return new String(m_object->GetString());
		}
		catch (const LuaPlus::LuaException& e)
		{
			throw new LuaException(new String(e.GetMessage()));
		}
	}
	
	String* GetWString()
	{
		try
		{
			return new String(m_object->GetWString());
		}
		catch (const LuaPlus::LuaException& e)
		{
			throw new LuaException(new String(e.GetMessage()));
		}
	}
	
	int StrLen()
	{
		try
		{
			return m_object->StrLen();
		}
		catch (const LuaPlus::LuaException& e)
		{
			throw new LuaException(new String(e.GetMessage()));
		}
	}
	
	lua_CFunction GetCFunction()
	{
		try
		{
			return m_object->GetCFunction();
		}
		catch (const LuaPlus::LuaException& e)
		{
			throw new LuaException(new String(e.GetMessage()));
		}
	}
	
	void* GetUserData()
	{
		try
		{
			return m_object->GetUserData();
		}
		catch (const LuaPlus::LuaException& e)
		{
			throw new LuaException(new String(e.GetMessage()));
		}
	}
	
	const void* GetLuaPointer()
	{
		try
		{
			return m_object->GetLuaPointer();
		}
		catch (const LuaPlus::LuaException& e)
		{
			throw new LuaException(new String(e.GetMessage()));
		}
	}

	void* GetLightUserData()
	{
		try
		{
			return m_object->GetLightUserData();
		}
		catch (const LuaPlus::LuaException& e)
		{
			throw new LuaException(new String(e.GetMessage()));
		}
	}

	bool GetBoolean()
	{
		try
		{
			return m_object->GetBoolean();
		}
		catch (const LuaPlus::LuaException& e)
		{
			throw new LuaException(new String(e.GetMessage()));
		}
	}

	LuaStackObject* PushStack();

//	LuaStackObject __cdecl PCall(const char* argTypes, ...);
//	LuaStackObject __cdecl PCall(int nresults, const char* argTypes, ...);

	void Call([ParamArray] Object* objs[]);

	LuaObject* GetMetaTable()			{  return new LuaObject(m_parent, m_object->GetMetaTable());  }
	void SetMetaTable(LuaObject* value)	{  m_object->SetMetaTable(*value->m_object);  }

	int GetN()
	{
		try
		{
			return m_object->GetN();
		}
		catch (const LuaPlus::LuaException& e)
		{
			throw new LuaException(new String(e.GetMessage()));
		}
	}

	void SetN(int n)
	{
		try
		{
			m_object->SetN(n);
		}
		catch (const LuaPlus::LuaException& e)
		{
			throw new LuaException(new String(e.GetMessage()));
		}
	}

	void Insert(LuaObject* obj)
	{
		try
		{
			m_object->Insert(*obj->m_object);
		}
		catch (const LuaPlus::LuaException& e)
		{
			throw new LuaException(new String(e.GetMessage()));
		}
	}

	void Insert(int index, LuaObject* obj)
	{
		try
		{
			m_object->Insert(index, *obj->m_object);
		}
		catch (const LuaPlus::LuaException& e)
		{
			throw new LuaException(new String(e.GetMessage()));
		}
	}

	void Remove(int index)
	{
		try
		{
			m_object->Remove(index);
		}
		catch (const LuaPlus::LuaException& e)
		{
			throw new LuaException(new String(e.GetMessage()));
		}
	}

	void Sort()
	{
		try
		{
			m_object->Sort();
		}
		catch (const LuaPlus::LuaException& e)
		{
			throw new LuaException(new String(e.GetMessage()));
		}
	}

	int GetCount()
	{
		try
		{
			return m_object->GetCount();
		}
		catch (const LuaPlus::LuaException& e)
		{
			throw new LuaException(new String(e.GetMessage()));
		}
	}

	int GetTableCount()
	{
		try
		{
			return m_object->GetTableCount();
		}
		catch (const LuaPlus::LuaException& e)
		{
			throw new LuaException(new String(e.GetMessage()));
		}
	}

	LuaObject* Clone()					{  return new LuaObject(m_parent, m_object->Clone());  }

	LuaObject* CreateTable(String* key)
		{  return new LuaObject(m_parent, m_object->CreateTable(MarshalString(key)));  }
	LuaObject* CreateTable(int key)
		{  return new LuaObject(m_parent, m_object->CreateTable(key));  }
	LuaObject* CreateTable(LuaObject* key)
		{  return new LuaObject(m_parent, m_object->CreateTable(*key->m_object));  }

	LuaObject* SetNil(String* key)
	{
		try
		{
			m_object->SetNil(MarshalString(key));
		}
		catch (const LuaPlus::LuaException& e)
		{
			throw new LuaException(new String(e.GetMessage()));
		}

		return this;
	}

	LuaObject* SetNil(int key)
	{
		try
		{
			m_object->SetNil(key);
		}
		catch (const LuaPlus::LuaException& e)
		{
			throw new LuaException(new String(e.GetMessage()));
		}

		return this;
	}

	LuaObject* SetNil(LuaObject* key)
	{
		try
		{
			m_object->SetNil(*key->m_object);
		}
		catch (const LuaPlus::LuaException& e)
		{
			throw new LuaException(new String(e.GetMessage()));
		}

		return this;
	}

	LuaObject* SetBoolean(String* key, bool value)
	{
		try
		{
			m_object->SetBoolean(MarshalString(key), value);
		}
		catch (const LuaPlus::LuaException& e)
		{
			throw new LuaException(new String(e.GetMessage()));
		}

		return this;
	}

	LuaObject* SetBoolean(int key, bool value)
	{
		try
		{
			m_object->SetBoolean(key, value);
		}
		catch (const LuaPlus::LuaException& e)
		{
			throw new LuaException(new String(e.GetMessage()));
		}

		return this;
	}

	LuaObject* SetBoolean(LuaObject* key, bool value)
	{
		try
		{
			m_object->SetBoolean(*key->m_object, value);
		}
		catch (const LuaPlus::LuaException& e)
		{
			throw new LuaException(new String(e.GetMessage()));
		}

		return this;
	}

	LuaObject* SetNumber(String* key, lua_Number value)
	{
		try
		{
			m_object->SetNumber(MarshalString(key), value);
		}
		catch (const LuaPlus::LuaException& e)
		{
			throw new LuaException(new String(e.GetMessage()));
		}

		return this;
	}

	LuaObject* SetNumber(int key, lua_Number value)
	{
		try
		{
			m_object->SetNumber(key, value);
		}
		catch (const LuaPlus::LuaException& e)
		{
			throw new LuaException(new String(e.GetMessage()));
		}

		return this;
	}

	LuaObject* SetNumber(LuaObject* key, lua_Number value)
	{
		try
		{
			m_object->SetNumber(*key->m_object, value);
		}
		catch (const LuaPlus::LuaException& e)
		{
			throw new LuaException(new String(e.GetMessage()));
		}

		return this;
	}

	LuaObject* SetString(String* key, String* value)
	{
		try
		{
			m_object->SetString(MarshalString(key), MarshalString(value));
		}
		catch (const LuaPlus::LuaException& e)
		{
			throw new LuaException(new String(e.GetMessage()));
		}

		return this;
	}

	LuaObject* SetString(int key, String* value)
	{
		try
		{
			m_object->SetString(key, MarshalString(value));
		}
		catch (const LuaPlus::LuaException& e)
		{
			throw new LuaException(new String(e.GetMessage()));
		}

		return this;
	}

	LuaObject* SetString(LuaObject* key, String* value)
	{
		try
		{
			m_object->SetString(*key->m_object, MarshalString(value));
		}
		catch (const LuaPlus::LuaException& e)
		{
			throw new LuaException(new String(e.GetMessage()));
		}

		return this;
	}

	LuaObject* SetWString(String* key, String* value)
	{
		try
		{
			m_object->SetWString(MarshalString(key), MarshalWString(value));
		}
		catch (const LuaPlus::LuaException& e)
		{
			throw new LuaException(new String(e.GetMessage()));
		}

		return this;
	}

	LuaObject* SetWString(int key, String* value)
	{
		try
		{
			m_object->SetWString(key, MarshalWString(value));
		}
		catch (const LuaPlus::LuaException& e)
		{
			throw new LuaException(new String(e.GetMessage()));
		}

		return this;
	}

	LuaObject* SetWString(LuaObject* key, String* value)
	{
		try
		{
			m_object->SetWString(*key->m_object, MarshalWString(value));
		}
		catch (const LuaPlus::LuaException& e)
		{
			throw new LuaException(new String(e.GetMessage()));
		}

		return this;
	}

	LuaObject* SetUserData(String* key, void* value)
	{
		try
		{
			m_object->SetUserData(MarshalString(key), value);
		}
		catch (const LuaPlus::LuaException& e)
		{
			throw new LuaException(new String(e.GetMessage()));
		}

		return this;
	}

	LuaObject* SetUserData(int key, void* value)
	{
		try
		{
			m_object->SetUserData(key, value);
		}
		catch (const LuaPlus::LuaException& e)
		{
			throw new LuaException(new String(e.GetMessage()));
		}

		return this;
	}

	LuaObject* SetUserData(LuaObject* key, void* value)
	{
		try
		{
			m_object->SetUserData(*key->m_object, value);
		}
		catch (const LuaPlus::LuaException& e)
		{
			throw new LuaException(new String(e.GetMessage()));
		}

		return this;
	}

	LuaObject* SetLightUserData(String* key, void* value)
	{
		try
		{
			m_object->SetLightUserData(MarshalString(key), value);
		}
		catch (const LuaPlus::LuaException& e)
		{
			throw new LuaException(new String(e.GetMessage()));
		}

		return this;
	}

	LuaObject* SetLightUserData(int key, void* value)
	{
		try
		{
			m_object->SetLightUserData(key, value);
		}
		catch (const LuaPlus::LuaException& e)
		{
			throw new LuaException(new String(e.GetMessage()));
		}

		return this;
	}

	LuaObject* SetLightUserData(LuaObject* key, void* value)
	{
		try
		{
			m_object->SetLightUserData(*key->m_object, value);
		}
		catch (const LuaPlus::LuaException& e)
		{
			throw new LuaException(new String(e.GetMessage()));
		}

		return this;
	}

	LuaObject* SetObject(String* key, LuaObject* value)
	{
		try
		{
			m_object->SetObject(MarshalString(key), *value->m_object);
		}
		catch (const LuaPlus::LuaException& e)
		{
			throw new LuaException(new String(e.GetMessage()));
		}

		return this;
	}

	LuaObject* SetObject(int key, LuaObject* value)
	{
		try
		{
			m_object->SetObject(key, *value->m_object);
		}
		catch (const LuaPlus::LuaException& e)
		{
			throw new LuaException(new String(e.GetMessage()));
		}

		return this;
	}

	LuaObject* SetObject(LuaObject* key, LuaObject* value)
	{
		try
		{
			m_object->SetObject(*key->m_object, *value->m_object);
		}
		catch (const LuaPlus::LuaException& e)
		{
			throw new LuaException(new String(e.GetMessage()));
		}

		return this;
	}


	void AssignNil(ManagedLuaPlus::LuaState* state);
	void AssignBoolean(ManagedLuaPlus::LuaState* state, bool value);
	void AssignNumber(ManagedLuaPlus::LuaState* state, lua_Number value);
	void AssignString(ManagedLuaPlus::LuaState* state, String* value);
	void AssignWString(ManagedLuaPlus::LuaState* state, String* value);
	void AssignUserData(ManagedLuaPlus::LuaState* state, void* value);
	void AssignLightUserData(ManagedLuaPlus::LuaState* state, void* value);
	void AssignObject(LuaObject* value);
	void AssignNewTable(ManagedLuaPlus::LuaState* state);

	LuaObject* GetByName(String* name)						{  return new LuaObject(m_parent, m_object->GetByName(MarshalString(name)));  }
	LuaObject* GetByIndex(int index)						{  return new LuaObject(m_parent, m_object->GetByIndex(index));  }
	LuaObject* GetByObject(LuaStackObject* obj);
	LuaObject* GetByObject(LuaObject* obj)					{  return new LuaObject(m_parent, m_object->GetByObject(*obj->m_object));  }

	__property LuaObject* get_Item(String* name)
	{
		try
		{
			return new LuaObject(m_parent, m_object->GetByName(MarshalString(name)));
		}
		catch (const LuaPlus::LuaException& e)
		{
			throw new LuaException(new String(e.GetMessage()));
		}
	}

	__property LuaObject* get_Item(int index)
	{
		try
		{
			return new LuaObject(m_parent, m_object->GetByIndex(index));
		}
		catch (const LuaPlus::LuaException& e)
		{
			throw new LuaException(new String(e.GetMessage()));
		}
	}

	__property LuaObject* get_Item(LuaObject* obj)
	{
		try
		{
			return new LuaObject(m_parent, m_object->GetByObject(*obj->m_object));
		}
		catch (const LuaPlus::LuaException& e)
		{
			throw new LuaException(new String(e.GetMessage()));
		}
	}

	LuaObject* Lookup(String* key)
	{
		return new LuaObject(m_parent, m_object->Lookup(MarshalString(key)));
	}

	void Register(String* funcName, LuaPlusCallback* function);
	void Register(String* funcName, Object* instance, MethodInfo* function);
	void Register(String* funcName, Delegate* function);

	__property String* get_WatchType()
	{
		int type = Type();
		
		switch (type)
		{
			case LuaPlus::LuaState::TNONE:
				return new String("NONE");
			
			case LuaPlus::LuaState::TNIL:
				return new String("NIL");

			case LuaPlus::LuaState::TBOOLEAN:
			{
				String* str = String::Format("BOOLEAN: {0}", GetBoolean() ? S"true" : S"false");
				return str;
			}

			case LuaPlus::LuaState::TLIGHTUSERDATA:
			{
				String* str = String::Format("LIGHTUSERDATA: {0:X}", __box((size_t)GetLightUserData()));
				return str;
			}

			case LuaPlus::LuaState::TNUMBER:
			{
				String* str = String::Format("NUMBER: {0:G}", __box(GetNumber()));
				return str;
			}

			case LuaPlus::LuaState::TSTRING:
			{
				String* str = String::Format("STRING: {0}", GetString());
				return str;
			}

			case LuaPlus::LuaState::TTABLE:
			{
				String* str = new String("TABLE");
				return str;
			}

			case LuaPlus::LuaState::TFUNCTION:
			{
				String* str = new String("FUNCTION");
				return str;
			}

			case LuaPlus::LuaState::TUSERDATA:
			{
				String* str = String::Format("USERDATA: {0:X}", __box((size_t)GetUserData()));
				return str;
			}

			case LuaPlus::LuaState::TTHREAD:
			{
				String* str = new String("FORMAT");
				return str;
			}

			case LuaPlus::LuaState::TWSTRING:
			{
				String* str = String::Format(L"WSTRING: {0}", GetWString());
				return str;
			}
		}

		return new String("Unknown");
	}

	__gc class TableEnumerator : public IEnumerator, public IDisposable
	{
	public:
		TableEnumerator(LuaObject* obj)
		{
			m_table = obj;
			if (!m_table->IsTable())
				throw new LuaException(new String("Object must be a table to enumerate over"));
			m_it = new LuaPlus::LuaTableIterator(*obj->m_object, false);
		}

		~TableEnumerator()
		{
			if (m_it)
				delete m_it;
		}

		void Reset() 
		{
			m_it->Reset();
		}

		bool MoveNext()
		{
			return m_it->Next();
		}

		__property Object* get_Current()
		{
			return new LuaNode(new LuaObject(m_table->m_parent, m_it->GetKey()), new LuaObject(m_table->m_parent, m_it->GetValue()));
		}

		// The current property on the IEnumerator interface:
/*		__property Object* IEnumerator::get_Current()
		{
			return (get_Current());
		}*/

		void Dispose()
		{
			delete m_it;
			m_it = NULL;
			m_table = NULL;
		}

	private:
		LuaPlus::LuaTableIterator __nogc* m_it;
		LuaObject* m_table;
	};

	IEnumerator* GetEnumerator() 
	{
		return new TableEnumerator(this);
	}

	// Implement the GetEnumerator() method:
	IEnumerator* IEnumerable::GetEnumerator() 
	{
		return GetEnumerator();
	}
	
public:
	LuaState* m_parent;
	LuaPlus::LuaObject __nogc* m_object;
};



public __gc class LuaStackObject
{
public:
	LuaStackObject(ManagedLuaPlus::LuaState* state, int index);
	LuaStackObject(ManagedLuaPlus::LuaState* state, LuaPlus::LuaStackObject obj);

	~LuaStackObject()
	{
		delete m_object;
	}

	__property String* get_WatchType()
	{
		int type = GetType();
		
		switch (type)
		{
			case LuaPlus::LuaState::TNONE:
				return new String("NONE");
			
			case LuaPlus::LuaState::TNIL:
				return new String("NIL");

			case LuaPlus::LuaState::TBOOLEAN:
			{
				String* str = String::Format("BOOLEAN: {0}", GetBoolean() ? S"true" : S"false");
				return str;
			}

			case LuaPlus::LuaState::TLIGHTUSERDATA:
			{
				String* str = String::Format("LIGHTUSERDATA: {0:X}", __box((size_t)GetLightUserData()));
				return str;
			}

			case LuaPlus::LuaState::TNUMBER:
			{
				String* str = String::Format("NUMBER: {0:G}", __box(GetNumber()));
				return str;
			}

			case LuaPlus::LuaState::TSTRING:
			{
				String* str = String::Format("STRING: {0}", GetString());
				return str;
			}

			case LuaPlus::LuaState::TTABLE:
			{
				String* str = new String("TABLE");
				return str;
			}

			case LuaPlus::LuaState::TFUNCTION:
			{
				String* str = new String("FUNCTION");
				return str;
			}

			case LuaPlus::LuaState::TUSERDATA:
			{
				String* str = String::Format("USERDATA: {0:X}", __box((size_t)GetUserData()));
				return str;
			}

			case LuaPlus::LuaState::TTHREAD:
			{
				String* str = new String("FORMAT");
				return str;
			}

			case LuaPlus::LuaState::TWSTRING:
			{
				String* str = String::Format(L"WSTRING: {0}", GetWString());
				return str;
			}
		}

		return new String("Unknown");
	}
	
	static int op_Implicit(ManagedLuaPlus::LuaStackObject* obj)
	{
		return (int)obj->m_object->m_stackIndex;
	}

	/**
		Various constructors accepting different parameters.
	**/
/*	LuaStackObject(lua_State* state, int stackIndex) :
		m_state(state),
		m_stackIndex(stackIndex)
	{
	}
*/
	/**
		Various constructors accepting different parameters.
	**/
//	LuaStackObject(LuaState* state, int stackIndex);

	/**
		Various constructors accepting different parameters.
	**/
//	LuaStackObject(LuaState& state, int stackIndex);

	/**
		Copy constructor.
	**/
/*	LuaStackObject(const LuaStackObject& src) :
		m_state(src.m_state),
		m_stackIndex(src.m_stackIndex)
	{
	}
*/
	/**
		Assignment operator.
	**/
/*	const LuaStackObject& operator=(const LuaStackObject& src)
	{
		m_state = src.m_state;
		m_stackIndex = src.m_stackIndex;
		return *this;
	}
*/
	/**
		Retrieves the LuaState object associated with this LuaStackObject.
	**/
	LuaState* GetLuaState()				{  return m_parent;  }

	String* GetTypeName()				{  return new String(m_object->GetTypeName());  }
	int GetType()						{  return m_object->GetType();  }

	bool IsNil()						{  return m_object->IsNil();  }
	bool IsTable()						{  return m_object->IsTable();  }
	bool IsUserData()					{  return m_object->IsUserData();  }
	bool IsCFunction()					{  return m_object->IsCFunction();  }
	bool IsNumber()						{  return m_object->IsNumber();  }
	bool IsString()						{  return m_object->IsString();  }
	bool IsWString()					{  return m_object->IsWString();  }
	bool IsFunction()					{  return m_object->IsFunction();  }
	bool IsNone()						{  return m_object->IsNone();  }
	bool IsLightUserData()				{  return m_object->IsLightUserData();  }
	bool IsBoolean()					{  return m_object->IsBoolean();  }

	int GetInteger()					{  return m_object->GetInteger();  }
	float GetFloat()					{  return m_object->GetFloat();  }
	double GetDouble()					{  return m_object->GetNumber();  }
	lua_Number GetNumber()				{  return m_object->GetNumber();  }
	String* GetString()					{  return new String(m_object->GetString());  }
	String* GetWString()				{  return new String(m_object->GetWString());  }
	int StrLen()						{  return m_object->StrLen();  }
	lua_CFunction GetCFunction()		{  return m_object->GetCFunction();  }
	void* GetUserData()					{  return m_object->GetUserData();  }
	const void* GetLuaPointer()			{  return m_object->GetLuaPointer();  }
	void* GetLightUserData()			{  return m_object->GetLightUserData();  }
	bool GetBoolean()					{  return m_object->GetBoolean();  }

	void Push()							{  return m_object->Push();  }
	void Pop()							{  return m_object->Pop();  }

	LuaStackObject* GetMetaTable()			{  return new LuaStackObject(m_parent, m_object->GetMetaTable());  }
	void SetMetaTable()					{  m_object->SetMetaTable();  }
	void SetMetaTable(LuaStackObject* value){  m_object->SetMetaTable(*value->m_object);  }

	void SetTable()						{  m_object->SetTable();  }
	int GetCount()						{  return m_object->GetCount();  }

	LuaStackObject* CreateTable(String* name)
		{  return new LuaStackObject(m_parent, m_object->CreateTable(MarshalString(name)));  }
	LuaStackObject* CreateTable(int index)
		{  return new LuaStackObject(m_parent, m_object->CreateTable(index));  }

	void SetNil(String* name)									{  m_object->SetNil(MarshalString(name));  }
	void SetNil(int index)										{  m_object->SetNil(index);  }
	void SetBoolean(String* name, bool value)					{  m_object->SetBoolean(MarshalString(name), value);  }
	void SetBoolean(int index, bool value)						{  m_object->SetBoolean(index, value);  }
	void SetNumber(String* name, double value)					{  m_object->SetNumber(MarshalString(name), (lua_Number)value);  }
	void SetNumber(int index, double value)						{  m_object->SetNumber(index, (lua_Number)value);  }
	void SetString(String* name, String* value)					{  m_object->SetString(MarshalString(name), MarshalString(value));  }
	void SetString(int index, String* value)					{  m_object->SetString(index, MarshalString(value));  }
	void SetWString(String* name, String* value)				{  m_object->SetWString(MarshalString(name), MarshalWString(value));  }
	void SetWString(int index, String* value)					{  m_object->SetWString(index, MarshalWString(value));  }
	void SetUserData(String* name, void* value)					{  m_object->SetUserData(MarshalString(name), value);  }
	void SetUserData(int index, void* value)					{  m_object->SetUserData(index, value);  }
	void SetLightUserData(String* name, void* value)			{  m_object->SetLightUserData(MarshalString(name), value);  }
	void SetObject(String* name, LuaStackObject* value)			{  m_object->SetObject(MarshalString(name), *value->m_object);  }
	void SetObject(int index, LuaStackObject* value)			{  m_object->SetObject(index, *value->m_object);  }

	LuaStackObject* GetByName(String* name)						{  return new LuaStackObject(m_parent, m_object->GetByName(MarshalString(name)));  }
	LuaStackObject* GetByIndex(int index)							{  return new LuaStackObject(m_parent, m_object->GetByIndex(index));  }
	LuaStackObject* GetByObject(LuaStackObject* obj)					{  return new LuaStackObject(m_parent, m_object->GetByObject(*obj->m_object));  }

public:
	LuaState* m_parent;
	LuaPlus::LuaStackObject __nogc* m_object;
};



public __gc class LuaState
{
public:
	__value enum ErrorCodes
	{
		/* error codes for `lua_load' and `lua_pcall' */
		ERRRUN = 1,
		ERRFILE = 2,
		ERRSYNTAX = 3,
		ERRMEM = 4,
		ERRERR = 5,
	};

	__value enum Types
	{
		TNONE = (-1),
		TNIL	= 0,
		TBOOLEAN	= 1,
		TLIGHTUSERDATA	= 2,
		TNUMBER = 3,
		TSTRING = 4,
		TTABLE = 5,
		TFUNCTION = 6,
		TUSERDATA = 7,
		TTHREAD = 8,
		TWSTRING = 9,
		NTYPES = 11,
	};

	__value enum DumpObjectTypes
	{
		DUMP_ALPHABETICAL		= 0x00000001,
		DUMP_WRITEALL			= 0x00000002,
		DUMP_WRITETABLEPOINTERS = 0x00000004,
	};

	LuaState();
	~LuaState();

	// LuaObject
	LuaObject* NewObject()						{  return new LuaObject(this);  }

	__property LuaObject* get_Globals()			{  return GetGlobals();  }
	LuaObject* GetGlobals()						{  return new LuaObject(this, m_state->GetGlobals());  }
	LuaObject* GetGlobal(String* name)			{  return new LuaObject(this, m_state->GetGlobal(MarshalString(name)));  }

	__property LuaObject* get_Registry()		{  return GetRegistry();  }
	LuaObject* GetRegistry()					{  return new LuaObject(this, m_state->GetRegistry());  }

	int Equal(const LuaObject* o1, const LuaObject* o2)
		{  return m_state->Equal(*o1->m_object, *o2->m_object);  }
	int LessThan(const LuaObject* o1, const LuaObject* o2)
		{  return m_state->LessThan(*o1->m_object, *o2->m_object);  }

	// Stack functions.
	LuaStackObject* Stack(int index)			{  return new LuaStackObject(this, index);  }
	int GetTop()								{  return m_state->GetTop();  }
	void SetTop(int index)						{  m_state->SetTop(index);  }
	void PushValue(int index)					{  m_state->PushValue(index);  }
	void PushValue(LuaStackObject* object)		{  m_state->PushValue(object->m_object->m_stackIndex);  }
	void Remove(int index)						{  m_state->Remove(index);  }
	void Insert(int index)						{  m_state->Insert(index);  }
	void Replace(int index)						{  m_state->Replace(index);  }
	int CheckStack(int size)					{  return m_state->CheckStack(size);  }
	void XMove(LuaState* to, int n)				{  m_state->XMove(to->m_state, n);  }
	int Equal(int index1, int index2)			{  return m_state->Equal(index1, index2);  }
	int RawEqual(int index1, int index2)		{  return m_state->RawEqual(index1, index2);  }
	int LessThan(int index1, int index2)		{  return m_state->LessThan(index1, index2);  }

	// push functions (C -> stack)
	void PushNil()								{  m_state->PushNil();    }
	void PushNumber(double n)					{  m_state->PushNumber((lua_Number)n);  }
	void PushLString(String* s, size_t len)		{  m_state->PushLString(MarshalString(s), len);  }
	void PushLWString(String* s, size_t len)	{  m_state->PushLWString(MarshalWString(s), len);  }
	void PushString(String* s)					{  m_state->PushString(MarshalString(s));  }
	void PushWString(String* s)					{  m_state->PushWString(MarshalWString(s));  }
//LUA_API const char *m_state->pushvfstring (State *L, const char *fmt,
//                                                    va_list argp);
//LUA_API const char *m_state->pushfstring (State *L, const char *fmt, ...);
//	void PushCClosure(CFunction fn, int n){  m_state->pushcclosure(fn, n);  }
//	void PushCClosure(LuaStateCFunction fn, int n)	{  m_state->pushcclosure((CFunction)fn, n);  }
//	void PushCFunction(CFunction f)		{  m_state->pushcclosure(f, 0);  }
//	void PushCFunction(LuaStateCFunction f)	{  m_state->pushcclosure((CFunction)f, 0);  }
	void PushBoolean(bool value)				{  m_state->PushBoolean(value);  }
	void PushLightUserData(void* p)				{  m_state->PushLightUserData(p);  }

	// get functions (Lua -> stack)
	void GetTable(int index)					{  m_state->GetTable(index);  }
	void RawGet(int index)						{  m_state->RawGet(index);  }
	void RawGetI(int index, int n)				{  m_state->RawGetI(index, n);  }
	LuaStackObject* NewTable(int /*size*/)		{  return new LuaStackObject(this, m_state->NewTable());  }
	void* NewUserData(size_t size)				{  return m_state->NewUserData(size);  }
	LuaStackObject* GetMetaTable(int index)		{  return new LuaStackObject(this, m_state->GetMetaTable(index));  }
	LuaStackObject* GetDefaultMetaTable(int type){  return new LuaStackObject(this, m_state->GetDefaultMetaTable(type));  }
	LuaStackObject* GetGlobals_Stack()			{  return new LuaStackObject(this, LUA_GLOBALSINDEX);  }
	LuaStackObject* GetGlobal_Stack(String* name)
	{
		LuaPlus::LuaStackObject obj = m_state->GetGlobal_Stack(MarshalString(name));
		return new LuaStackObject(this, obj); 
	}
	LuaStackObject* GetRegistry_Stack()			{  return new LuaStackObject(this, LUA_REGISTRYINDEX);  }
	LuaStackObject* GetRef(int ref)				{  return new LuaStackObject(this, m_state->GetRef(ref));  }

	// set functions(stack -> Lua)
	void SetTable(int index)					{  m_state->SetTable(index);  }
	void RawSet(int index)						{  m_state->RawSet(index);  }
	void RawSetI(int index, int n)				{  m_state->RawSetI(index, n);  }
	void SetMetaTable(int index)				{  m_state->SetMetaTable(index);  }
	void SetDefaultMetaTable(int type)			{  m_state->SetDefaultMetaTable(type);  }
	void SetGlobals(int index)					{  m_state->SetGlobals(index);  }
	void SetGlobal(String* name)				{  m_state->SetGlobal(MarshalString(name));  }
	int Ref(int lock)							{  return m_state->Ref(lock);  }
	void Unref(int ref)							{  m_state->Unref(ref);  }

	// `load' and `do' functions (load and run Lua code)
	void Call(int nargs, int nresults)			{  m_state->Call(nargs, nresults);  }
	int PCall(int nargs, int nresults, int errf){  return m_state->PCall(nargs, nresults, errf);  }
//	int CPCall(lua_CFunction func, void* ud)	{  return m_state->CPCall(func, ud);  }
//	int Dump(lua_Chunkwriter writer, void* data){  return m_state->Dump(writer, data);  }
	int LoadFile(String* filename)			{  return m_state->LoadFile(MarshalString(filename));  }
	int DoFile(String* filename)				{  return m_state->DoFile(MarshalString(filename));  }
	int DoString(String* str)					{  return m_state->DoString(MarshalString(str));  }
	int LoadBuffer(String* buff, size_t size, String* name)
	{
		return m_state->LoadBuffer(MarshalString(buff), size, MarshalString(name));
	}
	int DoBuffer(String* buff, size_t size, String* name)
	{
		return m_state->DoBuffer(MarshalString(buff), size, MarshalString(name));
	}

	// Coroutine functions
	int CoYield(int nresults)					{  return m_state->CoYield(nresults);  }
	int CoResume(int arg)						{  return m_state->CoResume(arg);  }

	// Garbage-collection functions
	int GetGCThreshold()						{  return m_state->GetGCThreshold();  }
	int GetGCCount()							{  return m_state->GetGCCount();  }
	void SetGCThreshold(int newthreshold)		{  m_state->SetGCThreshold(newthreshold);  }

	// Miscellaneous functions
	int Error()									{  return m_state->Error();  }

	int Next(int index)							{  return m_state->Next(index);  }
	int GetN(int index)							{  return m_state->GetN(index);  }

	void Concat(int n)							{  m_state->Concat(n);  }

	LuaObject* NewUserDataBox(void* u)			{  return new LuaObject(this, m_state->NewUserDataBox(u));  }
	LuaStackObject* NewUserDataBox_Stack(void* u){  return new LuaStackObject(this, m_state->NewUserDataBox_Stack(u));  }

	// Helper functions
	void Pop()									{  m_state->Pop(1);  }
	void Pop(int amount)						{  m_state->Pop(amount);  }

	void CollectGarbage()						{  m_state->SetGCThreshold(0);  }	

	bool DumpObject(String* filename, String* name, LuaObject* value)
	{
		return m_state->DumpObject(MarshalString(filename), MarshalString(name),
			*value->m_object, DUMP_ALPHABETICAL, 0, 0xffffffff); 
	}

	bool DumpObject(String* filename, String* name, LuaObject* value,
						 unsigned int flags, int indentLevel, unsigned int maxIndentLevel)
	{
		return m_state->DumpObject(MarshalString(filename), MarshalString(name),
			*value->m_object, flags, indentLevel, maxIndentLevel); 
	}

	bool DumpObject(String* filename, LuaObject* key, LuaObject* value,
						 unsigned int flags, int indentLevel, unsigned int maxIndentLevel)
	{
		return m_state->DumpObject(MarshalString(filename), *key->m_object,
			*value->m_object, flags, indentLevel, maxIndentLevel); 
	}

	bool DumpGlobals(String* filename, unsigned int flags, unsigned int maxIndentLevel)
	{
		return m_state->DumpGlobals(MarshalString(filename), flags, maxIndentLevel);
	}

	bool DumpGlobals(String* filename)
	{
		return m_state->DumpGlobals(MarshalString(filename), DUMP_ALPHABETICAL, 0xffffffff);
	}

	// Arg functions.
	int TypeError(int narg, const char* tname)							{  return m_state->TypeError(narg, tname);  }
	int ArgError(int narg, const char* extramsg)						{  return m_state->ArgError(narg, extramsg);  }
	const char* CheckLString(int numArg, size_t* len)					{  return m_state->CheckLString(numArg, len);  }
	const char* OptLString(int numArg, const char *def, size_t* len)	{  return m_state->OptLString(numArg, def, len);  }
	lua_Number CheckNumber(int numArg)									{  return m_state->CheckNumber(numArg);  }
	lua_Number OptNumber(int nArg, lua_Number def)						{  return m_state->OptNumber(nArg, def);  }
	void ArgCheck(bool condition, int numarg, const char* extramsg)		{  m_state->ArgCheck(condition, numarg, extramsg);  }
	const char* CheckString(int numArg)									{  return m_state->CheckString(numArg);  }
	const char* OptString(int numArg, const char* def)					{  return m_state->OptString(numArg, def);  }
	int CheckInt(int numArg)											{  return m_state->CheckInt(numArg);  }
	long CheckLong(int numArg)											{  return m_state->CheckLong(numArg);  }
	int OptInt(int numArg, int def)										{  return m_state->OptInt(numArg, def);  }
	long OptLong(int numArg, int def)									{  return m_state->OptLong(numArg, def);  }
	void CheckStack(int sz, const char* msg)							{  m_state->CheckStack(sz, msg);  }
	void CheckType(int narg, int t)										{  m_state->CheckType(narg, t);  }
	void CheckAny(int narg)												{  m_state->CheckAny(narg);  }
	void* CheckUData(int ud, const char* tname)							{  return m_state->CheckUData(ud, tname);  }

	// Debug functions.
/*	int GetStack(int level, Debug* ar)			{  return m_state->getstack(level, ar);  }
	int GetInfo(const char* what, Debug* ar)		{  return m_state->getinfo(what, ar);  }
	const char* GetLocal(const Debug* ar, int n)	{  return m_state->getlocal(ar, n);  }
	const char* SetLocal(const Debug* ar, int n)	{  return m_state->setlocal(ar, n);  }

	int SetHook(Hook func, unsigned long mask)	{  return m_state->sethook(func, mask);  }
	Hook GetHook()									{  return m_state->gethook(m_state);  }
	unsigned long GetHookMask()							{  return m_state->gethookmask(m_state);  }
*/
	LuaPlus::LuaState __nogc* m_state;

	// support for loader-lock issues
	// see http://support.microsoft.com/?id=814472
	static BOOL Initialize()
	{
		BOOL retval = TRUE;
		try
		{
			retval =  __crt_dll_initialize();
		}
		catch(System::Exception* e)
		{
			Console::WriteLine(e->Message);
			retval = FALSE;
		}
		return retval;
	}

	// support for loader-lock issues
	// see http://support.microsoft.com/?id=814472
	static BOOL Terminate()
	{
		BOOL retval = TRUE;
		try
		{
			retval = __crt_dll_terminate();
		}
		catch(System::Exception* e)
		{
			Console::WriteLine(e->Message);
			retval = FALSE;
		}
		return retval;
	}
};


public __gc class LuaTableIterator
{
public:
	LuaTableIterator(LuaObject* tableObj)
	{
		m_it = new LuaPlus::LuaTableIterator(*tableObj->m_object, true);
		m_tableObj = tableObj;
		if (!m_tableObj->IsTable())
			throw new LuaException(new String("Object must be a table to enumerate over"));
	}

	LuaTableIterator(LuaObject* tableObj, bool doReset)
	{
		m_it = new LuaPlus::LuaTableIterator(*tableObj->m_object, doReset);
		m_tableObj = tableObj;
		if (!m_tableObj->IsTable())
			throw new LuaException(new String("Object must be a table to enumerate over"));
	}

	~LuaTableIterator()
	{
		delete m_it;
	}

	void Reset()
	{
		m_it->Reset();
	}

	void Invalidate()
	{
		m_it->Invalidate();
	}

	bool Next()
	{
		return m_it->Next();
	}

	bool IsValid()
	{
		return m_it->IsValid();
	}

	static LuaTableIterator* op_Increment(LuaTableIterator* it)
	{
		it->Next();
		return it;
	}

	LuaObject* GetKey()
	{
		return new LuaObject(m_tableObj->m_parent, m_it->GetKey());
	}


	LuaObject* GetValue()
	{
		return new LuaObject(m_tableObj->m_parent, m_it->GetValue());
	}

protected:

private:
	LuaPlus::LuaTableIterator __nogc* m_it;
	LuaObject* m_tableObj;
};


///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
LuaStackObject::LuaStackObject(ManagedLuaPlus::LuaState* state, int index) :
	m_parent(state),
	m_object(new LuaPlus::LuaStackObject(state->m_state, index))
{
}


LuaStackObject::LuaStackObject(ManagedLuaPlus::LuaState* state, LuaPlus::LuaStackObject obj) :
	m_parent(state),
	m_object(new LuaPlus::LuaStackObject(obj))
{
}


LuaState::LuaState()
{
	m_state = LuaPlus::LuaState::Create();
}


LuaState::~LuaState()
{
	LuaPlus::LuaState::Destroy(m_state);
}


/**
	A helper class for automatically setting the stack state back to point of
	LuaAutoBlock creation.
**/
__gc class LuaAutoBlock
{
public:
	LuaAutoBlock(LuaState* state) :
		m_state(state)
	{
		m_stackTop = m_state->GetTop();
	}

	LuaAutoBlock(LuaStackObject* object) :
		m_state(object->m_parent)
	{
		m_stackTop = m_state->GetTop();
	}

	~LuaAutoBlock()
	{
		m_state->SetTop(m_stackTop);
	}

private:
	LuaState* m_state;
	int m_stackTop;
};


LuaObject::LuaObject(ManagedLuaPlus::LuaState* state) :
	m_object(new LuaPlus::LuaObject(state->m_state))
{
}

LuaObject::LuaObject(ManagedLuaPlus::LuaState* state, int stackIndex) :
	m_object(new LuaPlus::LuaObject(state->m_state, stackIndex))
{
}

LuaObject::LuaObject(ManagedLuaPlus::LuaObject* object) :
	m_object(new LuaPlus::LuaObject(*object->m_object))
{
}

LuaObject::LuaObject(ManagedLuaPlus::LuaState* parent, LuaPlus::LuaObject& object) :
	m_parent(parent),
	m_object(new LuaPlus::LuaObject(object))
{
}


LuaObject::LuaObject(ManagedLuaPlus::LuaStackObject* object) :
	m_object(new LuaPlus::LuaObject(object->m_object->m_state, *object->m_object))
{
}

	
LuaStackObject* LuaObject::PushStack()
{
	return new LuaStackObject(m_parent, m_object->PushStack());
}


LuaObject* LuaObject::GetByObject(LuaStackObject* obj)
{
	return new LuaObject(m_parent, m_object->GetByObject(*obj->m_object));
}

	
void LuaObject::AssignNil(ManagedLuaPlus::LuaState* state)
{
	try
	{
		m_object->AssignNil(state->m_state);
		m_parent = state;
	}
	catch (const LuaPlus::LuaException& e)
	{
		throw new LuaException(new String(e.GetMessage()));
	}
}

void LuaObject::AssignBoolean(ManagedLuaPlus::LuaState* state, bool value)
{
	try
	{
		m_object->AssignBoolean(state->m_state, value);
		m_parent = state;
	}
	catch (const LuaPlus::LuaException& e)
	{
		throw new LuaException(new String(e.GetMessage()));
	}
}

void LuaObject::AssignNumber(ManagedLuaPlus::LuaState* state, lua_Number value)
{
	try
	{
		m_object->AssignNumber(state->m_state, value);
		m_parent = state;
	}
	catch (const LuaPlus::LuaException& e)
	{
		throw new LuaException(new String(e.GetMessage()));
	}
}

void LuaObject::AssignString(ManagedLuaPlus::LuaState* state, String* value)
{
	try
	{
		m_object->AssignString(state->m_state, MarshalString(value));
		m_parent = state;
	}
	catch (const LuaPlus::LuaException& e)
	{
		throw new LuaException(new String(e.GetMessage()));
	}
}

void LuaObject::AssignWString(ManagedLuaPlus::LuaState* state, String* value)
{
	try
	{
		m_object->AssignWString(state->m_state, MarshalWString(value));
		m_parent = state;
	}
	catch (const LuaPlus::LuaException& e)
	{
		throw new LuaException(new String(e.GetMessage()));
	}
}

void LuaObject::AssignUserData(ManagedLuaPlus::LuaState* state, void* value)
{
	try
	{
		m_object->AssignUserData(state->m_state, value);
		m_parent = state;
	}
	catch (const LuaPlus::LuaException& e)
	{
		throw new LuaException(new String(e.GetMessage()));
	}
}

void LuaObject::AssignLightUserData(ManagedLuaPlus::LuaState* state, void* value)
{
	try
	{
		m_object->AssignLightUserData(state->m_state, value);
		m_parent = state;
	}
	catch (const LuaPlus::LuaException& e)
	{
		throw new LuaException(new String(e.GetMessage()));
	}
}

void LuaObject::AssignObject(LuaObject* value)
{
	try
	{
		m_object->AssignObject(*value->m_object);
		m_parent = value->m_parent;
	}
	catch (const LuaPlus::LuaException& e)
	{
		throw new LuaException(new String(e.GetMessage()));
	}
}

void LuaObject::AssignNewTable(ManagedLuaPlus::LuaState* state)
{
	try
	{
		m_object->AssignNewTable(state->m_state);
		m_parent = state;
	}
	catch (const LuaPlus::LuaException& e)
	{
		throw new LuaException(new String(e.GetMessage()));
	}
}


static int CallbackDispatcher(lua_State* inState)
{
	IntPtr callbackHandle = (IntPtr)lua_touserdata(inState, lua_upvalueindex(1));
	IntPtr stateHandle = (IntPtr)lua_touserdata(inState, lua_upvalueindex(2));
	ManagedLuaPlus::LuaPlusCallback* callback = __try_cast<ManagedLuaPlus::LuaPlusCallback*>(GCHandle::op_Explicit(callbackHandle).Target);
	ManagedLuaPlus::LuaState* state = __try_cast<ManagedLuaPlus::LuaState*>(GCHandle::op_Explicit(stateHandle).Target);
	return callback(state);
}


static int GCHandle_gc(lua_State* state)
{
	if (lua_isuserdata(state, 1))
	{
		GCHandle::op_Explicit((IntPtr)lua_touserdata(state, 1)).Free();
	}

	return 0;
}


LuaPlus::LuaObject CreateGCHandleObj(LuaPlus::LuaState* state, Object* theObj)
{
	LuaPlus::LuaObject metaTableObj = state->GetRegistry()["GCHandleMetaTable"];
	if (metaTableObj.IsNil())
	{
		metaTableObj = state->GetRegistry().CreateTable("GCHandleMetaTable");
		metaTableObj.Register("__gc", GCHandle_gc);
	}

	LuaPlus::LuaObject outObj = state->NewUserDataBox(GCHandle::op_Explicit(GCHandle::Alloc(theObj)).ToPointer());
	outObj.SetMetaTable(metaTableObj);

	return outObj;
}


void LuaObject::Register(String* funcName, LuaPlusCallback* function)
{
	PushStack();
	m_parent->PushString(funcName);

	LuaPlus::LuaObject functionHandleObj = CreateGCHandleObj(m_parent->m_state, function);
	LuaPlus::LuaObject parentObj = CreateGCHandleObj(m_parent->m_state, m_parent);
	functionHandleObj.PushStack();
	parentObj.PushStack();

	m_parent->m_state->PushCClosure(CallbackDispatcher, 2);

	m_parent->SetTable(-3);
	m_parent->Pop();
}


static int MethodInfoDispatcher(LuaPlus::LuaState* inState)
{
	IntPtr instanceHandle = (IntPtr)LuaPlus::LuaStackObject(inState, lua_upvalueindex(1)).GetUserData();
	IntPtr functionHandle = (IntPtr)LuaPlus::LuaStackObject(inState, lua_upvalueindex(2)).GetUserData();
	IntPtr stateHandle = (IntPtr)LuaPlus::LuaStackObject(inState, lua_upvalueindex(3)).GetUserData();
	Object* instance = __try_cast<Object*>(GCHandle::op_Explicit(instanceHandle).Target);
	MethodInfo* methodInfo = __try_cast<MethodInfo*>(GCHandle::op_Explicit(functionHandle).Target);
	ManagedLuaPlus::LuaState* state = __try_cast<ManagedLuaPlus::LuaState*>(GCHandle::op_Explicit(stateHandle).Target);

	ParameterInfo* parameters[] = methodInfo->GetParameters();

	int numParms = inState->GetTop();
	Object* parms[] = new Object*[numParms];
	for (int i = 1; i <= numParms; ++i)
	{
		LuaPlus::LuaStackObject obj(state->m_state, i);
		if (parameters[i - 1]->ParameterType == __typeof(int))
		{
			if (obj.IsNumber())
				parms[i - 1] = __box(obj.GetInteger());
		}
	}

	methodInfo->Invoke(instance, parms);
	return 0;
}


void LuaObject::Register(String* funcName, Object* instance, MethodInfo* function)
{
	PushStack();
	m_parent->PushString(funcName);

	LuaPlus::LuaObject instanceObj = CreateGCHandleObj(m_parent->m_state, instance);
	LuaPlus::LuaObject functionHandleObj = CreateGCHandleObj(m_parent->m_state, function);
	LuaPlus::LuaObject parentObj = CreateGCHandleObj(m_parent->m_state, m_parent);
	instanceObj.PushStack();
	functionHandleObj.PushStack();
	parentObj.PushStack();

	m_parent->m_state->PushCClosure(MethodInfoDispatcher, 3);

	m_parent->SetTable(-3);
	m_parent->Pop();
}


static int DelegateDispatcher(LuaPlus::LuaState* inState)
{
	IntPtr functionHandle = (IntPtr)LuaPlus::LuaStackObject(inState, lua_upvalueindex(1)).GetUserData();
	IntPtr stateHandle = (IntPtr)LuaPlus::LuaStackObject(inState, lua_upvalueindex(2)).GetUserData();
	Delegate* theDelegate = __try_cast<Delegate*>(GCHandle::op_Explicit(functionHandle).Target);
	ManagedLuaPlus::LuaState* state = __try_cast<ManagedLuaPlus::LuaState*>(GCHandle::op_Explicit(stateHandle).Target);

	MethodInfo* methodInfo = theDelegate->Method;
	ParameterInfo* parameters[] = methodInfo->GetParameters();

	int numParms = inState->GetTop();
	Object* parms[] = new Object*[numParms];
	for (int i = 1; i <= numParms; ++i)
	{
		LuaPlus::LuaStackObject obj(state->m_state, i);
		if (parameters[i - 1]->ParameterType == __typeof(int))
		{
			if (obj.IsNumber())
				parms[i - 1] = __box(obj.GetInteger());
		}
	}

	methodInfo->Invoke(theDelegate->Target, parms);
	return 0;
}


void LuaObject::Register(String* funcName, Delegate* function)
{
	PushStack();
	m_parent->PushString(funcName);

	LuaPlus::LuaObject functionHandleObj = CreateGCHandleObj(m_parent->m_state, function);
	LuaPlus::LuaObject parentObj = CreateGCHandleObj(m_parent->m_state, m_parent);
	functionHandleObj.PushStack();
	parentObj.PushStack();

	m_parent->m_state->PushCClosure(DelegateDispatcher, 2);

	m_parent->SetTable(-3);
	m_parent->Pop();
}


__gc private class Helper
{
public:
	static void Push(LuaState* state, int arg)
	{
		state->PushNumber(arg);
	}

	static void Push(LuaState* state, String* arg)
	{
		state->PushString(MarshalString(arg));
	}

	static void Push(LuaState* state, LuaObject* obj)
	{
		state;
		obj->PushStack();
	}

	static void Push(LuaState* state, LuaStackObject* obj)
	{
		state;
		obj->Push();
	}
};



void LuaObject::Call([ParamArray] Object* args[])
{
	PushStack();

	System::Type* types[] = new System::Type*[2];
	types[0] = __typeof(ManagedLuaPlus::LuaState);

	Object* objects[] = new Object*[2];
	objects[0] = m_parent;

	Helper* helper = new Helper();

	for (int i = 0; i < args->Length; ++i)
	{
		types[1] = args[i]->GetType();
		MethodInfo* pushInfo = __typeof(ManagedLuaPlus::Helper)->GetMethod("Push", types);
		if (!pushInfo)
		{
			String* str = String::Format("Array index {0} is not of a suitable type for pushing to the Lua stack.", __box(i));
			throw new System::ArgumentException(str);
		}

		objects[1] = args[i];
		pushInfo->Invoke(helper, objects);
	}

	m_parent->m_state->Call(args->Length, 0);
}


String* LuaNode::ToString()
{
	return String::Format("{0} / {1}", m_key->WatchType, m_value->WatchType);
}


} // namespace ManagedLuaPlus

