#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "LuaPlus/LuaPlus.h"
extern "C" {
#include "LuaPlus/include/lua.h"
#include "LuaPlus/include/lauxlib.h"
}
#include "luabeans.h"
#include "tLuaCOM.h"
#include "tLuaCOMException.h"

LuaBeans* LuaBeans::createBeans(LuaState* state, const char* name)
{
	if (state->GetGlobal("LuaBeansTable").IsNil())
	{
		state->GetGlobals().CreateTable("LuaBeansTable");
	}

	LuaBeans* lbeans = new LuaBeans(state, name);

	LuaObject luaBeansTableObj = state->GetGlobal("LuaBeansTable");
	luaBeansTableObj.SetUserData(name, lbeans);

	return lbeans;
}

LuaBeans* LuaBeans::getBeans(LuaState* state, const char* name)
{
	LuaObject luaBeansTableObj = state->GetGlobal("LuaBeansTable");
	LuaObject lbeansObj = luaBeansTableObj[name];
	LuaBeans* lbeans = (LuaBeans*)lbeansObj.GetUserData();

	return lbeans;
}

LuaBeans::LuaBeans(LuaState* state, const char* name) :
	m_state(state)
{
   sprintf(tag,"LB_%s_TAG",name);
   m_state->GetGlobals().CreateTable(tag);

   sprintf(ud_tag,"LB_%s_UDTAG",name);
   m_state->GetGlobals().CreateTable(ud_tag);
}

LuaBeans::~LuaBeans()
{
}

/**
**/
static int mm_newindex( LuaState* state )
{
	LuaStack args(state);
	LuaStackObject& tableObj = args[1];
	LuaStackObject& indexObj = args[2];
	LuaStackObject& valObj = args[3];

	try
	{
		tLuaCOM* lcom = (tLuaCOM*) tLuaCOM::lbeans->from_lua( tableObj );
		if ( indexObj.IsString() && lcom != NULL )
		{
			const char* name = indexObj.GetString();
			DISPID dispid;
			HRESULT hr = lcom->get_dispid(name, &dispid);

			if (!FAILED(hr))
			{
				FUNCDESC* pfuncdescr = NULL;
				hr = lcom->get_funcdescr( dispid, INVOKE_PROPERTYPUT, &pfuncdescr );
				if(FAILED(hr))
					hr = lcom->get_funcdescr( dispid, INVOKE_PROPERTYPUTREF, &pfuncdescr );
				if ( SUCCEEDED(hr) )
				{
					int result = lcom->putprop( pfuncdescr, valObj );
					lcom->ReleaseFuncDesc(pfuncdescr);
					if ( result > 0 )
						return 0;
				}
			}
		}
	}
	catch(class tLuaCOMException&)
	{
	}

	tableObj.SetObject( indexObj, valObj );
	return 0;
}


/*
* Closure que gerencia chamadas de metodos
*
* In: userdata (tLuaCOM *), userdata (FUNCDESC *),
*     table | userdata, ...
*/
static int lc_call(LuaState* state)
{
	tLuaCOM* lcom       = NULL;
	HRESULT hr          = S_OK;
	DISPID dispid       = -1;
	FUNCDESC *pFuncDesc = NULL;

	// posicao dos parametros
	const int luacom_obj_param  = 1;
	const int dispid_param      = 2;
	int user_param_pos          = 1; // precisa ser inicializado

	try
	{
		// obtem objeto luacom

		LuaStackObject obj(state, lua_upvalueindex(luacom_obj_param));
		lcom = (tLuaCOM *) obj.GetUserData();

		CHECK(lcom, INTERNAL_ERROR);

		// pega o dispid
		LuaStackObject dispidObj(state, lua_upvalueindex(dispid_param));
		dispid = (DISPID)dispidObj.GetUserData();

		// Testa para o caso de ser metodo
		hr = lcom->get_funcdescr(dispid,INVOKE_FUNC,&pFuncDesc);

		if (SUCCEEDED(hr))  //e' metodo
		{
			// Forca usuario a usar : na chamada a metodo, o que
			// faz com que lua passe o argumento "self". Esse argumento
			// nao e' utilizado na implementacao, mas e' exigido para
			// garantir uniformidade sintatica
			//      if(!check_self(LuaStackObject(state, 1)))
			//      {
			//        luaL_error(*state, "Method call without ':'");
			//        return 0;
			//      }

			int vals = lcom->call(pFuncDesc, tLuaObjList(state, 1));

			lcom->ReleaseFuncDesc(pFuncDesc);

			return vals; // fim do processo
		}    

		//// tenta para o caso de ser propget ou propput parametrizado

		// primeiro, pega a lista de parametros, para poder utilizar
		// o numero de parametros passados como discriminador entre
		// propget e propput

		tLuaObjList &params = tLuaObjList(state, 1);

		// testa o propget

		hr = lcom->get_funcdescr(dispid,INVOKE_PROPERTYGET,&pFuncDesc);    

		if(SUCCEEDED(hr) && (int)params.get_num_elements() >= pFuncDesc->cParams - pFuncDesc->cParamsOpt )
		{
			// e' propget

			int vals = lcom->call(pFuncDesc, params);

			lcom->ReleaseFuncDesc(pFuncDesc);

			return vals;
		}

		// tenta o propput (ou propputref) parametrizados

		hr = lcom->get_funcdescr(dispid,INVOKE_PROPERTYPUT,&pFuncDesc);

		if(FAILED(hr)) // tenta propputref
			hr = lcom->get_funcdescr(dispid,INVOKE_PROPERTYPUTREF,&pFuncDesc);

		if(SUCCEEDED(hr) && pFuncDesc->cParams == params.get_num_elements()) 
		{
			lcom->call(pFuncDesc, params);

			lcom->ReleaseFuncDesc(pFuncDesc);

			return 0;
		}
	}
	catch(class tLuaCOMException&)
	{
		// ignora excecoes
	}

	return 0;
}


/*
* tag method que gerencia leituras de campos
* do objeto luacom
*
* In: table, index
*/

static int mm_index( LuaState* state )
{
	LuaStack args(state);

	// variaveis utilizadas
	DISPID dispid;
	HRESULT hr              = S_OK;
	FUNCDESC* pfuncdescr    = NULL;
	tLuaCOM* lcom           = NULL;

	try
	{
		{
			lcom = (tLuaCOM*) tLuaCOM::lbeans->from_lua( args[1] );
			CHECK( lcom, INTERNAL_ERROR );
		}

		{
			LuaStackObject indexObj = args[2];

			CHECK( indexObj.IsString(), INTERNAL_ERROR );

			const char *name = indexObj.GetString();
			if (lcom->HasEnum())
			{
				if (!lcom->LookupEnum(name))
					state->PushNil();
				return 1;
			}
			else
			{
				hr = lcom->get_dispid(name, &dispid);
				if ( FAILED( hr ) ) 
				{
					// Wasn't an IDispatch function.  Could be an enum...
					tLuaCOM* newLCom = lcom->GetEnum(name);
					tLuaCOM::lbeans->push(newLCom); 
					return 1;
				}
			}
		}

		hr = lcom->get_funcdescr( dispid, INVOKE_PROPERTYGET, &pfuncdescr );
		if ( SUCCEEDED(hr) && pfuncdescr->cParams == 0 )
		{
			lcom->getprop(pfuncdescr);
		}
		else
		{
			state->NewUserDataBox_Stack( lcom );
			state->NewUserDataBox_Stack( (void *) dispid );
			state->PushCClosure( lc_call, 2 );
		}

		lcom->ReleaseFuncDesc(pfuncdescr);
	}
	catch(class tLuaCOMException&)
	{
		// ignora excecoes proprias
	}

	return 1;
}


/**
**/
static int mm_gc(LuaState* state)
{
	LuaStack args(state);
	if ( !args[1].IsUserData() )
		assert( 0 );		// How did we get here?
	tLuaCOM* lcom = (tLuaCOM*)args[1].GetUserData();
	delete lcom;

	return 0;
}


void LuaBeans::reg_tagmethods()
{
	LuaObject tagObj = m_state->GetGlobal(tag);
	tagObj.Register("__index", mm_index);
	tagObj.Register("__newindex", mm_newindex);
}

void LuaBeans::reg_ud_tagmethods()
{
	LuaObject tagObj = m_state->GetGlobal(ud_tag);
	tagObj.Register("__gc", mm_gc);
}

void LuaBeans::push( void* userdata )
{
	LuaObject table;
	table.AssignNewTable(m_state);

	LuaObject udMetaObj = m_state->GetGlobal(this->ud_tag);
	LuaObject dataObj = m_state->NewUserDataBox(userdata);
	dataObj.SetMetaTable(udMetaObj);
	table.SetObject("_USERDATA_REF_", dataObj);

	LuaObject tagObj = m_state->GetGlobal(this->tag);
	table.SetMetaTable(tagObj);

	table.PushStack();
}

void* LuaBeans::check_tag(int index)
{
   void* userdata = this->from_lua(LuaStackObject(m_state, index));

   luaL_argcheck(*m_state, (userdata!=NULL),index,"Object type is wrong");

   return userdata;
}

void* LuaBeans::from_lua(LuaObject luaobj)
{
	LuaObject tagObj = m_state->GetGlobal(tag);
	if (luaobj.IsTable()  &&  m_state->Equal(luaobj.GetMetaTable(), tagObj))
	{
		return luaobj["_USERDATA_REF_"].GetUserData();
	}
	else
		return NULL;
}

