/*
* tLuaCOM.cpp
*
*  Implementacao da classe tLuaCOM
*
* Renato Cerqueira
* Vinicius Almendra
*/

// RCS Info
static char *rcsid = "$Id: tLuaCOM.cpp,v 1.12 2001-04-30 09:56:07-03 almendra Exp almendra $";
static char *rcsname = "$Name:  $";

#include "tLuaCOM.h"
#include "tLuaDispatch.h"
#include "tUtil.h"
#include "tLuaCOMException.h"

char* tLuaCOM::TAG = "LuaCom";
LuaBeans* tLuaCOM::lbeans = NULL;

tLuaCOM::tLuaCOM(LuaState* state, IDispatch *pdisp_arg, ITypeInfo *ptinfo_arg) :
m_state(state)
{
	connection_point = NULL;
	connection_point_cookie = NULL;

	pdisp = pdisp_arg;
	if (pdisp)
		pdisp->AddRef();

	ptinfo = ptinfo_arg;
	ptinfo->AddRef();

	typehandler = new tLuaCOMTypeHandler(m_state, ptinfo_arg);
}

tLuaCOM::~tLuaCOM()
{
	if(connection_point != NULL)
		releaseConnection();

	delete typehandler;
	typehandler = NULL;

	ptinfo->Release();
	ptinfo = NULL;

	if (pdisp)
	{
		pdisp->Release();
		pdisp = NULL;
	}
}

HRESULT tLuaCOM::get_funcdescr(DISPID dispid, INVOKEKIND invkind,
							   FUNCDESC** ppfuncdesc)
{
	TYPEATTR * ptypeattr = NULL;
	HRESULT hr;
	unsigned int i, n;

	hr = ptinfo->GetTypeAttr(&ptypeattr);
	if (FAILED(hr))
	{
		return  hr;
	}

	n = ptypeattr->cFuncs;
	ptinfo->ReleaseTypeAttr(ptypeattr);

	for (i = 0; i < n;i++)
	{
		hr = ptinfo->GetFuncDesc(i, ppfuncdesc);
		if (FAILED(hr))
		{
			return  hr;
		}
		if (*ppfuncdesc == NULL)
		{
			return E_INVALIDARG;
		}

		//fazer o que tem que ser feito
		if ((*ppfuncdesc)->memid == dispid && (*ppfuncdesc)->invkind == invkind)
		{
			return S_OK;
		}

		ptinfo->ReleaseFuncDesc(*ppfuncdesc);
	}

	*ppfuncdesc = NULL;
	return E_INVALIDARG;
}

HRESULT tLuaCOM::get_dispid(const char* name, DISPID* dispid)
{
	HRESULT hr;
	wchar_t* w_name = (wchar_t*) malloc( (strlen(name) + 1) * sizeof(wchar_t));
	mbstowcs(w_name,name,strlen(name)+1);

	hr = pdisp->GetIDsOfNames(IID_NULL, &w_name, 1,
		LOCALE_SYSTEM_DEFAULT,dispid);
	free(w_name);
	return  hr;
}

int tLuaCOM::getprop(FUNCDESC* pfuncdesc)
{
	return call(pfuncdesc, tLuaObjList());
}

int tLuaCOM::call(FUNCDESC *pFuncDesc, tLuaObjList& params)
{
	HRESULT hr             = 0;
	unsigned int i         = 0;
	UINT ArgErr            = 0;

	DISPPARAMS dispparams;
	VARIANTARG result;
	EXCEPINFO excepinfo;

	VariantInit(&result);

	// nao suportado
	if (pFuncDesc->cParamsOpt == -1)
		return 0;

	// preenche estrutura com parametros
	typehandler->fillDispParams(dispparams, pFuncDesc, params);

	hr = pdisp->Invoke(
		pFuncDesc->memid,
		IID_NULL, 
		LOCALE_SYSTEM_DEFAULT,
		pFuncDesc->invkind,
		&dispparams,
		&result,
		&excepinfo,
		&ArgErr);

	if(SUCCEEDED(hr))
	{
		int numPushed = 1;

		/* empilha valor de retorno */
		typehandler->com2lua(result);

		if(pFuncDesc->invkind == INVOKE_FUNC)
			numPushed += typehandler->pushOutValues(params, pFuncDesc, dispparams);

		// Limpa parametros
		typehandler->releaseVariants(&dispparams);

		return numPushed;

	}
	else // Houve erro
	{
		// Limpa parametros
		typehandler->releaseVariants(&dispparams);

		if(hr == DISP_E_EXCEPTION) // excecoes
		{
			luaL_error(*m_state, (char *) 
				tUtil::bstr2string(excepinfo.bstrDescription));
		}
		else // Erros comuns
		{
			const char *errormessage = tUtil::GetErrorMessage(hr);

			if(errormessage != NULL)
			{
				luaL_error(*m_state, (char *) errormessage);
			}
			else
			{
				luaL_error(*m_state, "Unknown error");         
			}
		}

		return  0;
	}
}

int tLuaCOM::putprop(FUNCDESC* pfuncdesc, LuaStackObject val)
{
	int result;

	tLuaObjList params;
	params.addParam(val);

	result = call(pfuncdesc, params);

	return result;
}

bool tLuaCOM::addConnection(tLuaCOM *client)
{
	HRESULT hr;
	IDispatch *pdisp_client = client->GetIDispatch();

	IConnectionPointContainer *pcpc = NULL;

	assert(connection_point == NULL);
	if(connection_point != NULL)
		return false;

	hr = pdisp->QueryInterface
		(
		IID_IConnectionPointContainer, (void **) &pcpc
		);

	if(FAILED(hr))
	{
//		cout << "ERROR: Doesn't implement IConnectionPointContainer" << endl;
		return false;
	}

	{
		IID guid;

		client->GetIID(&guid);

		hr = pcpc->FindConnectionPoint(guid, &connection_point);
	}

	if(FAILED(hr))
	{
//		cout << "ERROR: Requested connection point not supported" << endl;

		pcpc->Release();

		pcpc = NULL;

		return false;
	}

	hr = connection_point->Advise
		(
		(IUnknown *) pdisp_client, &connection_point_cookie
		);

	if(FAILED(hr))
	{
//		cout << "ERROR: Unable to advise connection point" << endl;

		connection_point->Release();
		connection_point = NULL;

		pcpc->Release();
		pcpc = NULL;

		return false;
	}

	return true;
}

void tLuaCOM::releaseConnection(void)
{
	assert(connection_point);
	if(connection_point == NULL)
		return;

	connection_point->Unadvise(connection_point_cookie);
	connection_point->Release();
	connection_point = NULL;
}

#if 0
void tLuaCOM::getParamsInfo(const FUNCDESC * pfuncdesc,
							int *pTotal_params,
							int *pIn_params,
							int *pRequired_in_params)
{
	int total_params = 0;
	int in_params = 0;

	assert(pfuncdesc != NULL);

	if(!pfuncdesc)
		return;

	total_params = pfuncdesc->cParams;

	in_params = total_params;

	for (int i = 0; i < pfuncdesc->cParams; i++)
	{
		if ( !(pfuncdesc->lprgelemdescParam[i].idldesc.wIDLFlags & IDLFLAG_FIN) &&
			!(pfuncdesc->lprgelemdescParam[i].idldesc.wIDLFlags == IDLFLAG_NONE))
			in_params--;
	}

	if(pTotal_params != NULL)
		*pTotal_params = total_params;

	if(pIn_params != NULL)
		*pIn_params = in_params;

	if(pRequired_in_params != NULL)
		*pRequired_in_params = in_params - pfuncdesc->cParamsOpt;
}

/*
* getFuncDescInfo
*  Identifica o FUNCDESC correto
*/

void tLuaCOM::getFuncDescInfo(DISPID dispid, 
							  int num_real_params, 
							  FUNCDESC * * ppfuncdesc
							  )
{
	HRESULT hr = 0;

	hr = get_funcdescr(dispid,INVOKE_FUNC, ppfuncdesc);

	if(!FAILED(hr)) // E' um metodo
		return;

	hr = get_funcdescr(dispid,INVOKE_PROPERTYGET, ppfuncdesc);

	if(!FAILED(hr)) // Pode ser um propertyget
	{
		int inParams = 0;

		getParamsInfo(*ppfuncdesc, NULL, &inParams, NULL);

		if(inParams == num_real_params) // o numero de parametros bate com o get
			return;
		else
			ptinfo->ReleaseFuncDesc(*ppfuncdesc);
	}

	hr = get_funcdescr(dispid,INVOKE_PROPERTYPUT, ppfuncdesc);

	if(!FAILED(hr)) // Pode ser um propertyput
	{
		int total_params = 0;

		getParamsInfo(*ppfuncdesc, &total_params, NULL, NULL);

		if(total_params == num_real_params) // o numero de parametros bate com o put
			return;
		else
			ptinfo->ReleaseFuncDesc(*ppfuncdesc);
	}

	hr = get_funcdescr(dispid,INVOKE_PROPERTYPUTREF, ppfuncdesc);

	if(!FAILED(hr)) // Pode ser um propertyputref
	{
		int total_params = 0;

		getParamsInfo(*ppfuncdesc, &total_params, NULL, NULL);

		if(total_params == num_real_params) // o numero de parametros bate com o putref
			return;
		else
			ptinfo->ReleaseFuncDesc(*ppfuncdesc);
	}

	*ppfuncdesc = NULL; // nao e' nem metodo, nem property get, nem property put

	return;
}
#endif

#if 0
int tLuaCOM::check_and_call(DISPID dispid, lua_Object params_table)
{
	FUNCDESC *pfuncdesc = NULL;
	int retval = 0;
	int num_args = LuaAux::getNumEntries(params_table);

	getFuncDescInfo(
		dispid,
		num_args,    
		&pfuncdesc);

	if(pfuncdesc == NULL)
		luaL_error("Member not found");

	retval = call(pfuncdesc, params_table);

	ptinfo->ReleaseFuncDesc(pfuncdesc);

	return retval;
}
#endif

//
//  isMember
//
//    Informa se existe algum metodo ou propriedade com
//    o nome passado como parametro
//

bool tLuaCOM::isMember(const char * name)
{
	HRESULT hr;
	DISPID dumb_dispid;

	wchar_t* w_name = (wchar_t*) malloc( (strlen(name) + 1) * sizeof(wchar_t));

	assert(w_name);
	if(!w_name)
		return false;

	mbstowcs(w_name, name, strlen(name)+1);

	hr = pdisp->GetIDsOfNames(IID_NULL, &w_name, 1,
		LOCALE_SYSTEM_DEFAULT, &dumb_dispid);
	free(w_name);
	w_name = NULL;

	if(!FAILED(hr))
		return true;
	else
		return false;
}

bool tLuaCOM::GetDefaultEventsInterface(IUnknown *punk, 
										IID * piid, 
										ITypeInfo * * pptypeinfo)
{
	HRESULT hr    = S_OK;
	TYPEATTR *pTA = NULL;

	*piid=CLSID_NULL;

	if (!GetTypeInfoFromIUnknown(punk, pptypeinfo, SOURCE))
		return false;

	hr = (*pptypeinfo)->GetTypeAttr(&pTA);

	if (SUCCEEDED(hr))
	{
		*piid=pTA->guid;
		(*pptypeinfo)->ReleaseTypeAttr(pTA);
	}

	return true;
}

bool tLuaCOM::GetTypeInfoFromIUnknown(IUnknown *punk,
									  ITypeInfo * * pptypeinfo, 
									  tWhichInterface iface)
{
	HRESULT             hr = S_OK;
	LPTYPEINFO          pITypeInfoAll = NULL;
	LPTYPEATTR          pTA = NULL;

	assert(pptypeinfo);
	if (NULL == pptypeinfo)
		return false;

	if (!GetCoClassTypeInfo(punk, &pITypeInfoAll))
	{
		// tenta obter typeinfo correta a partir de IDispatch
		IDispatch *pdisp = NULL;

		hr = punk->QueryInterface(IID_IDispatch, (void **) &pdisp);

		if(FAILED(hr)) return false;

		hr = pdisp->GetTypeInfo(0, LOCALE_SYSTEM_DEFAULT, pptypeinfo);

		if(FAILED(hr))
		{
			pdisp->Release();
			return false;
		}

		TYPEATTR *ptypeattr = NULL;

		(*pptypeinfo)->GetTypeAttr(&ptypeattr);

		if(ptypeattr->typekind != TKIND_DISPATCH)
		{
			(*pptypeinfo)->ReleaseTypeAttr(ptypeattr);
			pdisp->Release();
			return false;
		}

		(*pptypeinfo)->ReleaseTypeAttr(ptypeattr);
		pdisp->Release();
		return true;
	}



	/*
	* We have the object's overall ITypeInfo in pITypeInfoAll.
	* Now get the type attributes which will tell us the number of
	* individual interfaces in this type library.  We then loop
	* through the "implementation types" for all those interfaces
	* calling GetImplTypeFlags, looking for the default.
	*/

	*pptypeinfo = NULL;  //Use this to determine success

	if (SUCCEEDED(pITypeInfoAll->GetTypeAttr(&pTA)))
	{
		UINT i;
		int iFlags;

		for (i=0; i < pTA->cImplTypes; i++)
		{
			//Get the implementation type for this interface
			hr = pITypeInfoAll->GetImplTypeFlags(i, &iFlags);

			if (FAILED(hr))
				continue;

			if (iFlags & IMPLTYPEFLAG_FDEFAULT || pTA->cImplTypes == 1)
			{
				if(iface == DISP && !(iFlags & IMPLTYPEFLAG_FSOURCE)
					|| iface == SOURCE && (iFlags & IMPLTYPEFLAG_FSOURCE))
				{
					HREFTYPE    hRefType=NULL;

					/*
					* This is the interface we want.  Get a handle to
					* the type description from which we can then get
					* the ITypeInfo.
					*/
					pITypeInfoAll->GetRefTypeOfImplType(i, &hRefType);
					hr = pITypeInfoAll->GetRefTypeInfo(hRefType
						, pptypeinfo);

					TYPEATTR *ptypeattr_iface = NULL;

					hr = (*pptypeinfo)->GetTypeAttr(&ptypeattr_iface);

					if(ptypeattr_iface->typekind == TKIND_INTERFACE)
					{
						(*pptypeinfo)->ReleaseTypeAttr(ptypeattr_iface);
						pITypeInfoAll->GetRefTypeOfImplType(-1, &hRefType);
						hr = pITypeInfoAll->GetRefTypeInfo(hRefType, pptypeinfo);
						hr = (*pptypeinfo)->GetTypeAttr(&ptypeattr_iface);
					}

					if(ptypeattr_iface->typekind == TKIND_DISPATCH)
					{
						(*pptypeinfo)->ReleaseTypeAttr(ptypeattr_iface);
						break;
					}
					else
					{
						(*pptypeinfo)->ReleaseTypeAttr(ptypeattr_iface);
						*pptypeinfo = NULL;
					}
				}
			}
		}

		pITypeInfoAll->ReleaseTypeAttr(pTA);
	}

	pITypeInfoAll->Release();

	if(*pptypeinfo != NULL)
		return true;
	else
	{
		// tenta obter typeinfo correta a partir de IDispatch
		IDispatch *pdisp = NULL;

		hr = punk->QueryInterface(IID_IDispatch, (void **) &pdisp);

		if(FAILED(hr)) return false;

		hr = pdisp->GetTypeInfo(0, LOCALE_SYSTEM_DEFAULT, pptypeinfo);

		if(FAILED(hr))
			return false;

		return true;
	}

}

bool tLuaCOM::GetCoClassTypeInfo(IUnknown *punk, ITypeInfo * * ppTypeinfo)
{
	IProvideClassInfo *pIProvideClassInfo = NULL;

	assert(ppTypeinfo);
	assert(punk);

	*ppTypeinfo=NULL;

	HRESULT hr = punk->QueryInterface(IID_IProvideClassInfo, (void **) &pIProvideClassInfo);

	if (FAILED(hr))
		return false;

	hr = pIProvideClassInfo->GetClassInfo(ppTypeinfo);
	pIProvideClassInfo->Release();

	return SUCCEEDED(hr);
}

void tLuaCOM::getHelpInfo(char **ppHelpFile, unsigned long *pHelpContext)
{
	ITypeLib *typelib = NULL;
	BSTR helpfile;
	HRESULT hr = S_OK;

	hr = ptinfo->GetDocumentation(-1, NULL, NULL, pHelpContext, &helpfile);

	if(FAILED(hr) || helpfile == NULL)
	{
		*ppHelpFile = NULL;
		return;
	}

	// Se nao conseguiu help contextna propria interface, tenta obte-lo
	// na type library
	if(*pHelpContext == 0)
	{
		unsigned int dumb_index = 0;
		unsigned long typelib_help_context = 0;
		BSTR helpfile_typelib;

		hr = ptinfo->GetContainingTypeLib(&typelib, &dumb_index);

		if(!FAILED(hr))
		{
			hr = typelib->GetDocumentation(-1, NULL, NULL,
				&typelib_help_context, &helpfile_typelib);

			if(!FAILED(hr))
			{
				SysFreeString(helpfile);

				helpfile = helpfile_typelib;
				*pHelpContext = typelib_help_context;
			}
		}
	}

	int str_size = SysStringLen(helpfile);
	*ppHelpFile = (char *) malloc((str_size + 1)*sizeof(wchar_t));
	wcstombs(*ppHelpFile, helpfile, str_size+1);

	SysFreeString(helpfile);
}



//
// CreateLuaCOM
//
//   Cria um objeto luacom a partir de um IUnknown,
//   Verificando a possibilidade de faze-lo
//
//

tLuaCOM *tLuaCOM::CreateLuaCOM(LuaState* state, IUnknown * punk)
{
	IDispatch *pdisp = NULL;
	ITypeInfo *ptypeinfo = NULL;
	IID default_iid;

	// obtem o IID e o TypeInfo da interface dispatch padrao
	bool result = GetDefaultDispinterface(punk, &default_iid, &ptypeinfo);

	if(result == false)
		return NULL;

	// obtem ponteiro para interface
	HRESULT hr = punk->QueryInterface(default_iid, (void **) &pdisp);

	if(FAILED(hr))
	{
		ptypeinfo->Release();
		return NULL;
	}

	tLuaCOM *pluacom = new tLuaCOM(state, pdisp, ptypeinfo);

	pdisp->Release();
	ptypeinfo->Release();

	return pluacom;
}

bool tLuaCOM::GetDefaultDispinterface(IUnknown *punk,
									  IID * piid,
									  ITypeInfo * * pptypeinfo)
{
	HRESULT hr    = S_OK;
	TYPEATTR *pTA = NULL;

	*piid=CLSID_NULL;

	bool result = GetTypeInfoFromIUnknown(punk, pptypeinfo, DISP);

	if(result == false)
		return false;

	hr = (*pptypeinfo)->GetTypeAttr(&pTA);

	if (FAILED(hr))
	{
		(*pptypeinfo)->Release();
		return false;
	}

	*piid = pTA->guid;
	(*pptypeinfo)->ReleaseTypeAttr(pTA);

	return true;
}


tLuaCOM * tLuaCOM::CreateLuaCOM(LuaState* state, IDispatch * pdisp)
{
	ITypeInfo *ptinfo = NULL;
	unsigned int is_there_typeinfo = 0;
	HRESULT hr;

	CHECK(pdisp, PARAMETER_OUT_OF_RANGE);

	hr = pdisp->GetTypeInfoCount(&is_there_typeinfo);

	if(FAILED(hr) || is_there_typeinfo == 0)
		return NULL;

	hr = pdisp->GetTypeInfo(0,LOCALE_SYSTEM_DEFAULT,&ptinfo);

	if(FAILED(hr))
		return NULL;

	tLuaCOM *lcom = new tLuaCOM(state, pdisp, ptinfo);

	return lcom;
}


tLuaCOM * tLuaCOM::CreateLuaCOM(LuaState* state, ITypeInfo * ptinfo)
{
	tLuaCOM *lcom = new tLuaCOM(state, NULL, ptinfo);

	return lcom;
}

ITypeInfo * tLuaCOM::GetDefaultEventsInterface()
{
	IID dumb_iid;
	ITypeInfo *ptinfo;

	bool result = GetDefaultEventsInterface(pdisp, &dumb_iid, &ptinfo);

	if(result == false)
		return NULL;

	return ptinfo;
}

void tLuaCOM::ReleaseFuncDesc(FUNCDESC * pfuncdesc)
{
	ptinfo->ReleaseFuncDesc(pfuncdesc);
}

IDispatch * tLuaCOM::GetIDispatch()
{
	return pdisp;
}

void tLuaCOM::GetIID(IID * piid)
{
	TYPEATTR *ptypeattr;

	ptinfo->GetTypeAttr(&ptypeattr);
	*piid = ptypeattr->guid;
	ptinfo->ReleaseTypeAttr(ptypeattr);
}


tLuaCOM* tLuaCOM::GetEnum(const char* name)
{
	ITypeLib *typelib = NULL;
	HRESULT hr;
	wchar_t* w_name = (wchar_t*) malloc( (strlen(name) + 1) * sizeof(wchar_t));
	mbstowcs(w_name,name,strlen(name)+1);

	UINT dummy;
	hr = ptinfo->GetContainingTypeLib(&typelib, &dummy);
    UINT infoCount(typelib->GetTypeInfoCount());
    for(UINT curEnum(0); curEnum < infoCount; ++curEnum) {
        TYPEKIND typeKind;
        hr = typelib->GetTypeInfoType(curEnum, &typeKind);
        if(hr || typeKind != TKIND_ENUM) continue;
        
		BSTR name;
        hr = typelib->GetDocumentation(curEnum, &name, NULL, 0, 0);
        if (FAILED(hr))
			continue;
		
		bool found = false;
		if (wcscmp(w_name, name) == 0)
			found = true;

		SysFreeString(name);

		if (found)
			break;
    }

	free(w_name);

	tLuaCOM* lcom = NULL;
	if (curEnum != infoCount)
	{
		ITypeInfo* outTypeInfo = NULL;
		hr = typelib->GetTypeInfo(curEnum, &outTypeInfo);
		if (SUCCEEDED(hr))
		{
			lcom = tLuaCOM::CreateLuaCOM(m_state, outTypeInfo);
			outTypeInfo->Release();
		}
	}

	return lcom;
}


bool tLuaCOM::LookupEnum(const char* name)
{
	wchar_t* w_name = (wchar_t*) malloc( (strlen(name) + 1) * sizeof(wchar_t));
	mbstowcs(w_name,name,strlen(name)+1);

	TYPEATTR* typeAttr;
	ptinfo->GetTypeAttr(&typeAttr);

	bool found = false;
	HRESULT hr = S_OK;
	for (UINT i = 0; i < typeAttr->cVars; ++i)
	{
		VARDESC* varDesc;
		ptinfo->GetVarDesc(i, &varDesc);

		BSTR name;
		UINT numNames;
		hr = ptinfo->GetNames(varDesc->memid, &name, 1, &numNames);

		if (wcscmp(name, w_name) == 0)
		{
			tLuaCOMTypeHandler comTypeHandler(m_state, ptinfo);
			comTypeHandler.com2lua(*varDesc->lpvarValue);

			found = true;
		}

		SysFreeString(name);

		ptinfo->ReleaseVarDesc(varDesc);

		if (found)
			break;
	}

	ptinfo->ReleaseTypeAttr(typeAttr);

	free(w_name);

	return found;
}


tLuaCOM* tLuaCOM::NewInterface(const char * pcInterface)
{
	HRESULT hr;

	wchar_t* wcInterface = (wchar_t*) 
		malloc( (strlen(pcInterface) + 1) * sizeof(wchar_t));
	mbstowcs(wcInterface, pcInterface, strlen(pcInterface)+1);

	const int max_typeinfos = 30;

	MEMBERID dumb[max_typeinfos];
	ITypeInfo *typeinfos[max_typeinfos];
	unsigned int number = 30;

	UINT dummy;
	ITypeLib* typelib;
	hr = ptinfo->GetContainingTypeLib(&typelib, &dummy);

	hr = typelib->FindName(wcInterface, 0, 
		typeinfos, dumb, (unsigned short *) &number);

	free(wcInterface);
	wcInterface = NULL;

	if(FAILED(hr) || number == 0)
		return false;

	unsigned int i = 0;
	TYPEATTR *typeattr;

	tLuaCOM* lcom = NULL;

	for(i = 0; i < number; i++)
	{
		hr = typeinfos[i]->GetTypeAttr(&typeattr);

		CHECK(SUCCEEDED(hr), INTERNAL_ERROR);

		if(typeattr->typekind == TKIND_COCLASS)
		{
			IDispatch* newDisp;
			hr = typeinfos[i]->CreateInstance(NULL, typeattr->guid, (void**)&newDisp);
			if (SUCCEEDED(hr))
			{
				lcom = tLuaCOM::CreateLuaCOM(m_state, newDisp);
				newDisp->Release();
				break;
			}
		}

		typeinfos[i]->ReleaseTypeAttr(typeattr);
	}

	return lcom;
}
