#include "../../LuaPlus/LuaPlus.h"
#ifdef WIN32
#include <windows.h>
#endif WIN32

using namespace LuaPlus;

#ifdef WIN32

struct FindFileInfo
{
	FindFileInfo() : m_handle(NULL) {}
	~FindFileInfo() {  CloseHandle( m_handle );  }
	HANDLE m_handle;
	WIN32_FIND_DATA m_fd;
};

#endif WIN32

static int LS_FindFirstFile( LuaState* state )
{
	LuaStack args(state);
	const char* wildcard = args[ 1 ].GetString();	wildcard;

#ifdef WIN32
	FindFileInfo* info = new FindFileInfo;
	info->m_handle = FindFirstFile(wildcard, &info->m_fd);
	if (info->m_handle == INVALID_HANDLE_VALUE)
	{
		delete info;
		return 0;
	}

	state->NewUserDataBox(info);
	state->GetGlobal("FileFind");
	state->SetMetaTable(-2);

	return 1;
#else !WIN32
	return 0;
#endif WIN32
}


static int LS_FindNextFile( LuaState* state )
{
	LuaStack args(state);
	if ( !args[ 1 ].IsUserData())
		return 0;

#ifdef WIN32
	FindFileInfo* info = static_cast<FindFileInfo*>( args[ 1 ].GetUserData() );
	BOOL ret = FindNextFile(info->m_handle, &info->m_fd);
	if (!ret)
	{
		FindClose(info->m_handle);
		info->m_handle = NULL;
		return 0;
	}

	state->PushBoolean(true);
	return 1;
#else !WIN32
	return 0;
#endif WIN32
}


static int LS_FindClose( LuaState* state )
{
	LuaStack args(state);
	if ( !args[ 1 ].IsUserData() )
		return 0;

#ifdef WIN32
	FindFileInfo* info = static_cast<FindFileInfo*>( args[ 1 ].GetUserData() );
	delete info;
#endif WIN32
	return 0;
}


static int LS_FindGetFileName( LuaState* state )
{
	LuaStack args(state);
	if ( !args[ 1 ].IsUserData() )
		return 0;

#ifdef WIN32
	FindFileInfo* info = static_cast<FindFileInfo*>( args[ 1 ].GetUserData() );
	state->PushString(info->m_fd.cFileName);
#endif WIN32

	return 1;
}


static int LS_FindIsDirectory( LuaState* state )
{
	LuaStack args(state);
	if ( !args[ 1 ].IsUserData() )
		return 0;

#ifdef WIN32
	FindFileInfo* info = static_cast<FindFileInfo*>( args[ 1 ].GetUserData() );
	state->PushBoolean((info->m_fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) != 0);

	return 1;
#else !WIN32
	return 0;
#endif WIN32
}


extern "C" __declspec(dllexport) void LuaPlus_RegisterModule( LuaState* state )
{
	LuaObject fileFindObj = state->GetGlobals().CreateTable("FileFind");
	fileFindObj.Register("First", LS_FindFirstFile);
	fileFindObj.Register("Next", LS_FindNextFile);
	fileFindObj.Register("Close", LS_FindClose);
	fileFindObj.Register("GetFileName", LS_FindGetFileName);
	fileFindObj.Register("IsDirectory", LS_FindIsDirectory);
}


extern "C" __declspec(dllexport) void LuaPlus_UnregisterModule( LuaState* state )
{
	state->GetGlobals().SetNil("FileFind");
}

